"""Implementation fo Zemel's Learned Fair Representations lifted from AIF360.

https://github.com/IBM/AIF360/blob/master/aif360/algorithms/preprocessing/lfr.py
"""
from typing import List, Tuple, Optional, Union

import numpy as np
import pandas as pd
import scipy.optimize as optim

from ethicml.implementations.utils import load_data_from_flags, save_transformations, PreAlgoArgs
from ethicml.utility.data_structures import DataTuple, TestTuple


# Disable pylint's naming convention complaints - this code wasn't implemented by us
# pylint: disable=invalid-name


class ZemelArgs(PreAlgoArgs):
    """Arguments for the Zemel algorithm."""

    clusters: int
    Ax: float
    Ay: float
    Az: float
    max_iter: int
    maxfun: int
    epsilon: float
    threshold: float


def distances(x: np.ndarray, v: np.ndarray, alpha: np.ndarray) -> np.ndarray:
    """Compute distances.

    Compute the l2 distance between each feature, x_n, and each
    prototype vector, v_k, weighted by a feature-wise weight parameter, alpha
    Args:
        x: Input features
        v: Prototype vectors
        alpha: Individual weight parameters for each feature dimension

    Returns: array of shape [n, k]containing the weighted l2 distances between
    each feature, x_n, and each prototype vector, v_k
    """
    dists = np.sum(np.square(x[:, None] - v[None]) * alpha, axis=-1)
    return dists


def softmax(dists: np.ndarray) -> np.ndarray:
    """Compute the probability that x_n maps to a given prototype vector, v_k.

    Args:
        dists: l2 distances between each input feature and each prototype vector

    Returns:
        Membership probabilities of each input feature to each prototype
    """
    exp = np.exp(-dists)
    denom = np.sum(exp, axis=1, keepdims=True)
    denom[denom == 0] = 1e-6
    matrix_nk = exp / denom

    return matrix_nk


def x_n_hat(x, matrix_nk, v) -> Tuple[np.ndarray, float]:
    """Reconstruct x_n from z_n.

    And compute the difference between the original input and its reconstruction using l2 loss.

    Args:
        x: Input features
        matrix_nk: Membership probabilities of each input feature to each prototype
        v: Prototype vectors

    Returns:
        Reconstruction of x and the l2 loss between the reconstruction and x
    """
    x_n_hat_obj = matrix_nk @ v
    l_x = (np.square(x - x_n_hat_obj)).sum()

    return x_n_hat_obj, l_x


def yhat(matrix_nk: np.ndarray, y: np.ndarray, w: np.ndarray) -> Tuple[np.ndarray, float]:
    """Predict y and compute the cross-entropy loss.

     This is between the resulting predictions and the target labels

    Args:
        matrix_nk: Matrix encoding the membership probabilities of each input feature to
        each prototype
        y: target labels for each sample
        w: learned parameters governing the mapping from the prototypes to classification decisions

    Returns:
        Predictions and associated cross-entropy loss
    """
    y_hat = yhat_without_loss(matrix_nk, w)
    l_y = (-y * np.log(y_hat) - (1.0 - y) * np.log(1.0 - y_hat)).sum()

    return y_hat, l_y


def yhat_without_loss(matrix_nk: np.ndarray, w: np.ndarray) -> np.ndarray:
    """Predict y without computing the classification loss.

    Args:
        matrix_nk: Matrix encoding the membership probabilities of each input feature to
        each prototype
        w: learned parameters governing the mapping from the prototypes to classification decisions

    Returns:
        Predictions and associated cross-entropy loss
    """
    y_hat = matrix_nk @ w
    y_hat[y_hat <= 0] = 1e-6
    y_hat[y_hat >= 1] = 0.999

    return y_hat


def lfr_optim_ob(
    params: np.ndarray,
    data_sensitive: np.ndarray,
    data_nonsensitive: np.ndarray,
    y_sensitive: np.ndarray,
    y_nonsensitive: np.ndarray,
    k: int = 10,
    a_x: float = 0.01,
    a_y: float = 0.1,
    a_z: float = 0.5,
    results: bool = False,
) -> Union[Tuple[np.ndarray, ...], float]:
    """Apply L-BFGS to minimize the LFR objective function.

    Args:
        params: Array of trainable parameters.
        data_sensitive: Data for the subset of individuals that are members
        of the protected set (i.e. S= 1)
        data_nonsensitive: Data for the subset of individuals that are not members
        of the protected set (i.e. S= 0)
        y_sensitive:
        y_nonsensitive:
        k: Number of prototype sets
        a_x: Pre-factor for the L_x loss term.
        a_y: Pre-factor for the L_y loss term.
        a_z: Pre-factor for the L_z loss term.
        results: Whether to return the predictions and prototype membership probabilities (True)
        or the weighted LFR loss (False)

    Returns:
        If 'results' is True, returns the predictions and prototype membership probabilities;
        otherwise returns the weighted LFR loss
    """
    lfr_optim_ob.iters += 1  # type: ignore[attr-defined]
    _, p = data_sensitive.shape

    alpha0 = params[:p]
    alpha1 = params[p : 2 * p]
    w = params[2 * p : (2 * p) + k]
    v = np.array(params[(2 * p) + k :]).reshape((k, p))

    dists_sensitive = distances(data_sensitive, v, alpha1)
    dists_nonsensitive = distances(data_nonsensitive, v, alpha0)

    m_nk_sensitive = softmax(dists_sensitive)
    m_nk_nonsensitive = softmax(dists_nonsensitive)

    m_k_sensitive = np.mean(m_nk_sensitive, axis=0)
    m_k_nonsensitive = np.mean(m_nk_nonsensitive, axis=0)

    # Loss term enforcing group fairness (minimizes MI between z and s)
    l_z = np.sum(np.abs(m_k_sensitive - m_k_nonsensitive))

    _, l_x1 = x_n_hat(data_sensitive, m_nk_sensitive, v)
    _, l_x2 = x_n_hat(data_nonsensitive, m_nk_nonsensitive, v)
    l_x = l_x1 + l_x2

    yhat_sensitive, l_y1 = yhat(m_nk_sensitive, y_sensitive, w)
    yhat_nonsensitive, l_y2 = yhat(m_nk_nonsensitive, y_nonsensitive, w)
    l_y = l_y1 + l_y2

    criterion = a_x * l_x + a_y * l_y + a_z * l_z

    return_tuple = (yhat_sensitive, yhat_nonsensitive, m_nk_sensitive, m_nk_nonsensitive)

    return return_tuple if results else criterion


lfr_optim_ob.iters = 0  # type: ignore[attr-defined]


def train_and_transform(
    train: DataTuple, test: TestTuple, flags: ZemelArgs
) -> (Tuple[DataTuple, TestTuple]):
    """Train the Zemel model and return the transformed features of the train and test sets."""
    np.random.seed(888)
    features_dim = train.x.shape[1]

    sens_col = train.s.columns[0]
    training_sensitive = train.x.loc[train.s[sens_col] == 0].to_numpy()
    training_nonsensitive = train.x.loc[train.s[sens_col] == 1].to_numpy()
    ytrain_sensitive = train.y.loc[train.s[sens_col] == 0].to_numpy()
    ytrain_nonsensitive = train.y.loc[train.s[sens_col] == 1].to_numpy()

    model_inits = np.random.uniform(
        size=int(features_dim * 2 + flags.clusters + features_dim * flags.clusters)
    )
    bnd: List[Tuple[Optional[int], Optional[int]]] = []
    for i, _ in enumerate(model_inits):
        if i < features_dim * 2 or i >= features_dim * 2 + flags.clusters:
            bnd.append((None, None))
        else:
            bnd.append((0, 1))

    learned_model = optim.fmin_l_bfgs_b(
        lfr_optim_ob,
        x0=model_inits,
        epsilon=flags.epsilon,
        args=(
            training_sensitive,
            training_nonsensitive,
            ytrain_sensitive[:, 0],
            ytrain_nonsensitive[:, 0],
            flags.clusters,
            flags.Ax,
            flags.Ay,
            flags.Az,
            0,
        ),
        bounds=bnd,
        approx_grad=True,
        maxfun=flags.maxfun,
        maxiter=flags.max_iter,
        disp=False,
    )[0]

    testing_sensitive = test.x.loc[test.s[sens_col] == 0].to_numpy()
    testing_nonsensitive = test.x.loc[test.s[sens_col] == 1].to_numpy()

    # Mutated, fairer dataset with new labels
    test_transformed = transform(
        testing_sensitive, testing_nonsensitive, learned_model, test, flags
    )

    training_sensitive = train.x.loc[train.s[sens_col] == 0].to_numpy()
    training_nonsensitive = train.x.loc[train.s[sens_col] == 1].to_numpy()
    ytrain_sensitive = train.y.loc[train.s[sens_col] == 0].to_numpy()
    ytrain_nonsensitive = train.y.loc[train.s[sens_col] == 1].to_numpy()

    # extract training model parameters
    train_transformed = transform(
        training_sensitive, training_nonsensitive, learned_model, train, flags
    )

    return (
        DataTuple(x=train_transformed, s=train.s, y=train.y, name=train.name),
        TestTuple(x=test_transformed, s=test.s, name=test.name),
    )


def transform(features_sens, features_nonsens, learned_model, dataset, flags: ZemelArgs):
    """Transform a dataset.

    Args:
        features_sens: Sensitive features.
        features_nonsens: Nonsensitive features.
        label_sens: Sensitive labels.
        label_nonsens: Class labels.
        learned_model: Model optimized for the LFR objective.
        dataset: Dataset to be transformed.

    Returns:
        Dataframe of transformed features.
    """
    k = flags.clusters
    _, p = features_sens.shape
    alphaoptim0 = learned_model[:p]
    alphaoptim1 = learned_model[p : 2 * p]
    voptim = np.array(learned_model[(2 * p) + k :]).reshape((k, p))

    # compute distances on the test dataset using train model params
    dist_sensitive = distances(features_sens, voptim, alphaoptim1)
    dist_nonsensitive = distances(features_nonsens, voptim, alphaoptim0)

    # compute cluster probabilities for test instances
    m_nk_sensitive = softmax(dist_sensitive)
    m_nk_nonsensitive = softmax(dist_nonsensitive)

    # learned mappings for test instances
    res_sensitive = x_n_hat(features_sens, m_nk_sensitive, voptim)
    x_n_hat_sensitive = res_sensitive[0]
    res_nonsensitive = x_n_hat(features_nonsens, m_nk_nonsensitive, voptim)
    x_n_hat_nonsensitive = res_nonsensitive[0]

    sens_col = dataset.s.columns[0]

    sensitive_idx = dataset.x[dataset.s[sens_col] == 0].index
    nonsensitive_idx = dataset.x[dataset.s[sens_col] == 1].index

    transformed_features = np.zeros_like(dataset.x.to_numpy())
    transformed_features[sensitive_idx] = x_n_hat_sensitive
    transformed_features[nonsensitive_idx] = x_n_hat_nonsensitive

    return pd.DataFrame(transformed_features, columns=dataset.x.columns)


def main():
    """Main method to run model."""
    args = ZemelArgs()
    args.parse_args()

    train, test = load_data_from_flags(args)
    save_transformations(train_and_transform(train, test, args), args)


if __name__ == "__main__":
    main()



"""Wrapper for SKLearn implementation of SVM."""
from typing import Optional, Union

from sklearn.svm import SVC, LinearSVC
import pandas as pd

from ethicml.common import implements
from ethicml.utility import DataTuple, TestTuple
from .in_algorithm import InAlgorithm


class SVM(InAlgorithm):
    """Support Vector Machine."""

    def __init__(self, C: Optional[float] = None, kernel: Optional[str] = None):
        """Init SVM."""
        super().__init__()
        self.C = SVC().C if C is None else C
        self.kernel = SVC().kernel if kernel is None else kernel

    @implements(InAlgorithm)
    def run(self, train: DataTuple, test: Union[DataTuple, TestTuple]) -> pd.DataFrame:
        clf = select_svm(self.C, self.kernel)
        clf.fit(train.x, train.y.to_numpy().ravel())
        return pd.DataFrame(clf.predict(test.x), columns=["preds"])

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return "SVM"


def select_svm(C: float, kernel: str) -> SVC:
    """Select the appropriate SVM model for the given parameters."""
    print(kernel)
    if kernel == "linear":
        return LinearSVC(C=C, dual=False, tol=1e-12, random_state=888)
    return SVC(C=C, kernel=kernel, gamma="auto", random_state=888)



"""Class to describe features of the Adult dataset."""

from .dataset import Dataset


class Credit(Dataset):
    """UCL Credit Card dataset."""

    def __init__(self, split: str = "Sex", discrete_only: bool = False):
        """Init credit dataset."""
        super().__init__()
        self.split = split
        self.discrete_only = discrete_only
        self.features = [
            "LIMIT_BAL",
            "SEX",
            "EDUCATION_1",
            "EDUCATION_2",
            "EDUCATION_3",
            "MARRIAGE_1",
            "MARRIAGE_2",
            "MARRIAGE_3",
            "AGE",
            "PAY_0",
            "PAY_2",
            "PAY_3",
            "PAY_4",
            "PAY_5",
            "PAY_6",
            "BILL_AMT1",
            "BILL_AMT2",
            "BILL_AMT3",
            "BILL_AMT4",
            "BILL_AMT5",
            "BILL_AMT6",
            "PAY_AMT1",
            "PAY_AMT2",
            "PAY_AMT3",
            "PAY_AMT4",
            "PAY_AMT5",
            "PAY_AMT6",
            "default-payment-next-month",
        ]

        self.continuous_features = [
            "LIMIT_BAL",
            "AGE",
            "PAY_0",
            "PAY_2",
            "PAY_3",
            "PAY_4",
            "PAY_5",
            "PAY_6",
            "BILL_AMT1",
            "BILL_AMT2",
            "BILL_AMT3",
            "BILL_AMT4",
            "BILL_AMT5",
            "BILL_AMT6",
            "PAY_AMT1",
            "PAY_AMT2",
            "PAY_AMT3",
            "PAY_AMT4",
            "PAY_AMT5",
            "PAY_AMT6",
        ]

        if split == "Sex":
            self.sens_attrs = ["SEX"]
            self.s_prefix = ["SEX"]
            self.class_labels = ["default-payment-next-month"]
            self.class_label_prefix = ["default-payment-next-month"]
        elif split == "Custom":
            pass
        else:
            raise NotImplementedError

    @property
    def name(self) -> str:
        """Getter for dataset name."""
        return f"Credit {self.split}"

    @property
    def filename(self) -> str:
        """Getter for file name."""
        return "UCI_Credit_Card.csv"



"""Implementation of Beutel's adversarially learned fair representations."""
# Disable pylint checking overwritten method signatures. Pytorch forward passes use **kwargs
# pylint: disable=arguments-differ

import random
from typing import List, Any, Tuple, Sequence
import pandas as pd
import numpy as np
import torch
from torch import nn, Tensor
from torch.autograd import Function
from torch.optim.lr_scheduler import ExponentialLR
from torch.optim import Adam

from ethicml.utility.data_structures import DataTuple, TestTuple, FairnessType
from ethicml.implementations.utils import load_data_from_flags, save_transformations, PreAlgoArgs
from ethicml.preprocessing.train_test_split import train_test_split
from ethicml.preprocessing.adjust_labels import assert_binary_labels, LabelBinarizer
from .pytorch_common import CustomDataset, TestDataset

STRING_TO_ACTIVATION_MAP = {"Sigmoid()": nn.Sigmoid()}

STRING_TO_LOSS_MAP = {"BCELoss()": nn.BCELoss()}


class BeutelArgs(PreAlgoArgs):
    """Args for the Beutel Implementation."""

    fairness: FairnessType
    enc_size: List[int]
    adv_size: List[int]
    pred_size: List[int]
    enc_activation: str
    adv_activation: str
    batch_size: int
    y_loss: str
    s_loss: str
    epochs: int
    adv_weight: float
    validation_pcnt: float


def set_seed(seed: int):
    """Set the seeds for numpy torch etc."""
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)  # type: ignore[attr-defined]


def build_networks(flags: BeutelArgs, train_data: CustomDataset, enc_activation, adv_activation):
    """Build the networks we use.

    Pulled into a separate function to make the code a bit neater.
    """
    enc = Encoder(
        enc_size=flags.enc_size, init_size=int(train_data.xdim), activation=enc_activation
    )
    adv = Adversary(
        fairness=flags.fairness,
        adv_size=flags.adv_size,
        init_size=flags.enc_size[-1],
        s_size=int(train_data.sdim),
        activation=adv_activation,
        adv_weight=flags.adv_weight,
    )
    pred = Predictor(
        pred_size=flags.pred_size,
        init_size=flags.enc_size[-1],
        class_label_size=int(train_data.ydim),
        activation=adv_activation,
    )

    model = Model(enc, adv, pred)

    return enc, model


def make_dataset_and_loader(
    data: DataTuple, flags: BeutelArgs
) -> Tuple[CustomDataset, torch.utils.data.DataLoader]:
    """Given a datatuple, create a dataset and a corresponding dataloader.

    Args:
        data:
        flags:

    Returns:
        Tuple of a pytorch dataset and dataloader.
    """
    dataset = CustomDataset(data)
    dataloader = torch.utils.data.DataLoader(
        dataset=dataset, batch_size=flags.batch_size, shuffle=False
    )
    return dataset, dataloader


def train_and_transform(
    train: DataTuple, test: TestTuple, flags: BeutelArgs
) -> Tuple[DataTuple, TestTuple]:
    """Train the fair autoencoder on the training data and then transform both training and test."""
    set_seed(888)

    post_process = False
    if flags.y_loss == "BCELoss()":
        try:
            assert_binary_labels(train)
        except AssertionError:
            processor = LabelBinarizer()
            train = processor.adjust(train)
            post_process = True

    # By default we use 10% of the training data for validation
    train_, validation = train_test_split(train, train_percentage=1 - flags.validation_pcnt)

    train_data, train_loader = make_dataset_and_loader(train_, flags)
    _, validation_loader = make_dataset_and_loader(validation, flags)
    _, all_train_data_loader = make_dataset_and_loader(train, flags)

    test_data = TestDataset(test)
    test_loader = torch.utils.data.DataLoader(
        dataset=test_data, batch_size=flags.batch_size, shuffle=False
    )

    # convert flags to Python objects
    enc_activation = STRING_TO_ACTIVATION_MAP[flags.enc_activation]
    adv_activation = STRING_TO_ACTIVATION_MAP[flags.adv_activation]
    y_loss_fn = STRING_TO_LOSS_MAP[flags.y_loss]
    s_loss_fn = STRING_TO_LOSS_MAP[flags.s_loss]

    enc, model = build_networks(
        flags=flags,
        train_data=train_data,
        enc_activation=enc_activation,
        adv_activation=adv_activation,
    )

    optimizer = torch.optim.Adam(model.parameters())
    scheduler = ExponentialLR(optimizer, gamma=0.95)

    best_val_loss = torch.ones(1) * np.inf
    best_enc = None

    for i in range(1, flags.epochs + 1):
        model.train()
        for embedding, sens_label, class_label in train_loader:
            _, s_pred, y_pred = model(embedding, class_label)

            loss = y_loss_fn(y_pred, class_label)

            if flags.fairness == "EqOp":
                mask = class_label.ge(0.5)
            elif flags.fairness == "EqOd":
                raise NotImplementedError("Not implemented Eq. Odds yet")
            elif flags.fairness == "DP":
                mask = torch.ones(s_pred.shape, dtype=torch.uint8)
            loss += s_loss_fn(
                s_pred, torch.masked_select(sens_label, mask).view(-1, int(train_data.sdim))
            )

            step(i, loss, optimizer, scheduler)

        if i % 5 == 0 or i == flags.epochs:
            model.eval()
            val_y_loss = torch.zeros(1)
            val_s_loss = torch.zeros(1)
            for embedding, sens_label, class_label in validation_loader:
                _, s_pred, y_pred = model(embedding, class_label)

                val_y_loss += y_loss_fn(y_pred, class_label)

                mask = get_mask(flags, s_pred, class_label)

                val_s_loss -= s_loss_fn(
                    s_pred, torch.masked_select(sens_label, mask).view(-1, int(train_data.sdim))
                )

            val_loss = (val_y_loss / len(validation_loader)) + (val_s_loss / len(validation_loader))

            if val_loss < best_val_loss:
                best_val_loss = val_loss
                best_enc = enc.state_dict()

    enc.load_state_dict(best_enc)

    transformed_train = encode_dataset(enc, all_train_data_loader, train)
    if post_process:
        transformed_train = processor.post(encode_dataset(enc, all_train_data_loader, train))
    return transformed_train, encode_testset(enc, test_loader, test)


def step(iteration: int, loss: Tensor, optimizer: Adam, scheduler: ExponentialLR) -> None:
    """Do one training step."""
    optimizer.zero_grad()
    loss.backward()
    optimizer.step()
    scheduler.step(iteration)


def get_mask(flags: BeutelArgs, s_pred, class_label):
    """Get a mask to enforce different fairness types."""
    if flags.fairness == "EqOp":
        mask = class_label.ge(0.5)
    elif flags.fairness == "EqOd":
        raise NotImplementedError("Not implemented Eq. Odds yet")
    elif flags.fairness == "DP":
        mask = torch.ones(s_pred.shape, dtype=torch.uint8)

    return mask


def encode_dataset(
    enc: nn.Module, dataloader: torch.utils.data.DataLoader, datatuple: DataTuple
) -> DataTuple:
    """Encode a dataset."""
    data_to_return: List[Any] = []

    for embedding, _, _ in dataloader:
        data_to_return += enc(embedding).data.numpy().tolist()

    return DataTuple(
        x=pd.DataFrame(data_to_return),
        s=datatuple.s,
        y=datatuple.y,
        name=f"Beutel: {datatuple.name}",
    )


def encode_testset(
    enc: nn.Module, dataloader: torch.utils.data.DataLoader, testtuple: TestTuple
) -> TestTuple:
    """Encode a dataset.

    Args:
        enc:
        dataloader:

    Returns:
        Encoded TestTuple.
    """
    data_to_return: List[Any] = []

    for embedding, _ in dataloader:
        data_to_return += enc(embedding).data.numpy().tolist()

    return TestTuple(x=pd.DataFrame(data_to_return), s=testtuple.s, name=testtuple.name)


class GradReverse(Function):
    """Gradient reversal layer."""

    @staticmethod
    def forward(ctx, x: Tensor, lambda_: float) -> Tensor:  # type: ignore[override]
        """Forward pass."""
        ctx.lambda_ = lambda_
        return x.view_as(x)

    @staticmethod
    def backward(ctx, grad_output):
        """Backward pass with Gradient reversed / inverted."""
        return grad_output.neg().mul(ctx.lambda_), None


def _grad_reverse(features: Tensor, lambda_: float) -> Tensor:
    return GradReverse.apply(features, lambda_)  # type: ignore[attr-defined]


class Encoder(nn.Module):
    """Encoder of the GAN."""

    def __init__(self, enc_size: Sequence[int], init_size: int, activation):
        """Init Encoder."""
        super().__init__()
        self.encoder = nn.Sequential()
        if not enc_size:  # In the case that encoder size [] is specified
            self.encoder.add_module("single encoder layer", nn.Linear(init_size, init_size))
            self.encoder.add_module("single layer encoder activation", activation)
        else:
            self.encoder.add_module("encoder layer 0", nn.Linear(init_size, enc_size[0]))
            self.encoder.add_module("encoder activation 0", activation)
            for k in range(len(enc_size) - 1):
                self.encoder.add_module(
                    "encoder layer {}".format(k + 1), nn.Linear(enc_size[k], enc_size[k + 1])
                )
                self.encoder.add_module("encoder activation {}".format(k + 1), activation)

    def forward(self, x: Tensor) -> Tensor:  # type: ignore[override]
        """Forward pass."""
        encoded = self.encoder(x)
        return encoded


class Adversary(nn.Module):
    """Adversary of the GAN."""

    def __init__(
        self,
        fairness: FairnessType,
        adv_size: Sequence[int],
        init_size: int,
        s_size: int,
        activation: nn.Module,
        adv_weight: float,
    ):
        """Init Adversary."""
        super().__init__()
        self.fairness = fairness
        self.init_size = init_size
        self.adv_weight = adv_weight
        self.adversary = nn.Sequential()
        if not adv_size:  # In the case that encoder size [] is specified
            self.adversary.add_module("single adversary layer", nn.Linear(init_size, s_size))
            self.adversary.add_module("single layer adversary activation", activation)
        else:
            self.adversary.add_module("adversary layer 0", nn.Linear(init_size, adv_size[0]))
            self.adversary.add_module("adversary activation 0", activation)
            for k in range(len(adv_size) - 1):
                self.adversary.add_module(
                    "adversary layer {}".format(k + 1), nn.Linear(adv_size[k], adv_size[k + 1])
                )
                self.adversary.add_module("adversary activation {}".format(k + 1), activation)
            self.adversary.add_module("adversary last layer", nn.Linear(adv_size[-1], s_size))
            self.adversary.add_module("adversary last activation", activation)

    def forward(self, x: Tensor, y: Tensor) -> Tensor:  # type: ignore[override]
        """Forward pass."""
        x = _grad_reverse(x, lambda_=self.adv_weight)

        if self.fairness == "EqOp":
            mask = y.ge(0.5)
            x = torch.masked_select(x, mask).view(-1, self.init_size)
            x = self.adversary(x)
        elif self.fairness == "EqOd":
            raise NotImplementedError("Not implemented equalized odds yet")
        elif self.fairness == "DP":
            x = self.adversary(x)
        return x


class Predictor(nn.Module):
    """Predictor of the GAN."""

    def __init__(
        self, pred_size: Sequence[int], init_size: int, class_label_size: int, activation: nn.Module
    ):
        """Init predictor class."""
        super().__init__()
        self.predictor = nn.Sequential()
        if not pred_size:  # In the case that encoder size [] is specified
            self.predictor.add_module(
                "single adversary layer", nn.Linear(init_size, class_label_size)
            )
            self.predictor.add_module("single layer adversary activation", activation)
        else:
            self.predictor.add_module("adversary layer 0", nn.Linear(init_size, pred_size[0]))
            self.predictor.add_module("adversary activation 0", activation)
            for k in range(len(pred_size) - 1):
                self.predictor.add_module(
                    "adversary layer {}".format(k + 1), nn.Linear(pred_size[k], pred_size[k + 1])
                )
                self.predictor.add_module("adversary activation {}".format(k + 1), activation)
            self.predictor.add_module(
                "adversary last layer", nn.Linear(pred_size[-1], class_label_size)
            )
            self.predictor.add_module("adversary last activation", activation)

    def forward(self, x: Tensor) -> Tensor:  # type: ignore[override]
        """Forward pass."""
        return self.predictor(x)


class Model(nn.Module):
    """Whole GAN model."""

    def __init__(self, enc: Encoder, adv: Adversary, pred: Predictor) -> None:
        """Init Model object."""
        super().__init__()
        self.enc = enc
        self.adv = adv
        self.pred = pred

    def forward(  # type: ignore[override]
        self, x: Tensor, y: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """Forward pass."""
        encoded = self.enc(x)
        s_hat = self.adv(encoded, y)
        y_hat = self.pred(encoded)
        return encoded, s_hat, y_hat


def main():
    """Load data from feather files, pass it to `train_and_transform` and then save the result."""
    args = BeutelArgs().parse_args()
    train, test = load_data_from_flags(args)
    save_transformations(train_and_transform(train, test, args), args)


if __name__ == "__main__":
    main()



"""This module contains items related to data, such as raw csv's and data objects."""

from .adult import Adult
from .compas import Compas
from .configurable_dataset import ConfigurableDataset
from .credit import Credit
from .dataset import filter_features_by_prefixes, Dataset
from .german import German
from .load import load_data, create_data_obj
from .non_binary_toy import NonBinaryToy
from .sqf import Sqf
from .toy import Toy



"""Loads Data from .csv files."""

from pathlib import Path
from typing import List, Dict, Optional
import pandas as pd

from ethicml.utility.data_structures import DataTuple
from .configurable_dataset import ConfigurableDataset
from .dataset import Dataset


def load_data(dataset: Dataset, ordered: bool = False) -> DataTuple:
    """Load dataset from its CSV file.

    Args:
        dataset: dataset object
        ordered: if True return features such that discrete come first, then continuous

    Returns:
        DataTuple with dataframes of features, labels and sensitive attributes
    """
    dataframe: pd.DataFrame = pd.read_csv(dataset.filepath)
    assert isinstance(dataframe, pd.DataFrame)

    feature_split = dataset.feature_split if not ordered else dataset.ordered_features

    x_data = dataframe[feature_split["x"]]
    s_data = dataframe[feature_split["s"]]
    y_data = dataframe[feature_split["y"]]

    return DataTuple(x=x_data, s=s_data, y=y_data, name=dataset.name)


def create_data_obj(
    filepath: Path,
    s_columns: List[str],
    y_columns: List[str],
    additional_to_drop: Optional[List[str]] = None,
) -> ConfigurableDataset:
    """Create a `ConfigurableDataset` from the given file.

    Args:
        filepath: path to a CSV file
        s_columns: list of columns that represent sensitive attributes
        y_columns: list of columns that contain lables
        additional_to_drop: other columns that should be dropped

    Returns:
        Dataset object
    """
    if additional_to_drop is None:
        additional_to_drop = []

    conf: ConfigurableDataset = ConfigurableDataset()
    conf.filename = filepath.name
    conf.filepath = filepath

    dataframe: pd.DataFrame = pd.read_csv(filepath)

    columns: List[str] = dataframe.columns.to_numpy().tolist()
    for s_col in s_columns:
        columns.remove(s_col)
    for y_col in y_columns:
        columns.remove(y_col)
    for additional in additional_to_drop:
        columns.remove(additional)

    feat_split: Dict[str, List[str]] = {"x": columns, "s": s_columns, "y": y_columns}
    conf.feature_split = feat_split

    return conf



"""Test helper functions"""
import pytest

from ethicml.common import implements


def test_implements() -> None:
    """test the implements decorator"""

    class BaseClass:
        def func(self) -> None:
            """Do nothing."""

        def no_docstring(self) -> None:
            pass

    class CorrectImplementation(BaseClass):
        @implements(BaseClass)
        def func(self) -> None:
            pass

    with pytest.raises(AssertionError):

        class IncorrectImplementation(BaseClass):
            @implements(BaseClass)
            def wrong_func(self) -> None:
                pass

    with pytest.raises(AssertionError):

        class NoDocstringImpl(BaseClass):
            @implements(BaseClass)
            def no_docstring(self) -> None:
                pass



"""Wrapper for calling the fair GP model."""
import sys
from pathlib import Path
from tempfile import TemporaryDirectory

from typing import List, Optional, Dict, Any
import numpy as np
import pandas as pd

from ethicml.common import implements
from ethicml.utility import DataTuple, TestTuple
from .in_algorithm import InAlgorithmAsync
from .installed_model import InstalledModel

REPO_URL = "https://github.com/predictive-analytics-lab/fair-gpytorch.git"
PRED_FNAME = "predictions.npz"
MAX_EPOCHS = 1000
MAX_BATCH_SIZE = 10100  # can go up to 10000
MAX_NUM_INDUCING = 5000  # 2500 seems to be more than enough
SEED = 888


class GPyT(InstalledModel):
    """Normal GP model."""

    basename = "GPyT"

    def __init__(self, s_as_input=True, flags=None, code_dir=None):
        """Instantiate the model.

        Args:
            s_as_input: if True, use s as an input feature
            flags: (optional) a dictionary that can overwrite *any* flag
            code_dir: (optional) a directory where the GPyT code can be found. if this is not given,
                      then the code is downloaded from GitHub
        """
        if code_dir is None:
            super().__init__(dir_name="gpyt", top_dir="fair-gpytorch", url=REPO_URL)
        else:
            super().__init__(dir_name=str(code_dir), top_dir="", executable=sys.executable)
        self.s_as_input = s_as_input
        self.flag_overwrites: Dict[str, Any] = {} if flags is None else flags

    @implements(InAlgorithmAsync)
    async def run_async(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        (ytrain,), label_converter = _fix_labels([train.y.to_numpy()])
        raw_data = dict(
            xtrain=train.x.to_numpy(),
            strain=train.s.to_numpy(),
            ytrain=ytrain,
            xtest=test.x.to_numpy(),
            stest=test.s.to_numpy(),
            # ytest=ytest,
        )
        parameters = self._additional_parameters(raw_data)

        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            data_path = tmp_path / "data.npz"
            np.savez(data_path, **raw_data)
            model_name = "local"
            flags = _flags(
                parameters,
                str(data_path),
                tmpdir,
                self.s_as_input,
                model_name,
                raw_data["ytrain"].shape[0],
            )
            await self._run_gpyt(flags)

            # Read the results from the numpy file 'predictions.npz'
            with (tmp_path / model_name / PRED_FNAME).open("rb") as file_obj:
                output = np.load(file_obj)
                pred_mean = output["pred_mean"]

        predictions = label_converter((pred_mean > 0.5).astype(raw_data["ytrain"].dtype)[:, 0])
        return pd.DataFrame(predictions, columns=["preds"])

    async def _run_gpyt(self, flags):
        """Generate command to run GPyT."""
        cmd = [str(self._code_path / "run.py")]
        # apply flag overwrites
        cmd_flags = {**flags, **self.flag_overwrites}
        for key, value in cmd_flags.items():
            cmd += [f"--{key}", str(value)]
        await self._call_script(cmd)

    @staticmethod
    def _additional_parameters(_):
        return dict(lik="BaselineLikelihood")

    @property
    def name(self):
        """Getter for the algorithm name."""
        return f"{self.basename}_in_{self.s_as_input}"


class GPyTDemPar(GPyT):
    """GP algorithm which enforces demographic parity."""

    MEAN = 2  # pylint: disable=invalid-name
    MIN = 3  # pylint: disable=invalid-name
    MAX = 4  # pylint: disable=invalid-name

    def __init__(
        self,
        target_acceptance=None,
        average_prediction=False,
        target_mode=MEAN,
        marginal=False,
        precision_target=1.0,
        **kwargs,
    ):
        """Instantiate the model.

        Args:
            s_as_input: should the sensitive attribute be part of the input?
            target_acceptance: which acceptance rate to target
            average_prediction: whether to use to average of all possible sensitive attributes for
                                predictions
            target_mode: if no target rate is given, how is the target chosen?
            marginal: when doing average_prediction, should the prior of s be taken into account?
            precision_target: how similar should target labels and true labels be
        """
        super().__init__(**kwargs)
        if self.s_as_input and average_prediction:
            self.__name = f"{self.basename}_dem_par_av_True"
            if marginal:
                self.__name += "_marg"
        else:
            self.__name = f"{self.basename}_dem_par_in_{self.s_as_input}"
        if target_acceptance is not None:
            self.__name += f"_tar_{target_acceptance}"
        elif target_mode != self.MEAN:
            if target_mode == self.MIN:
                self.__name += "_tar_min"
            elif target_mode == self.MAX:
                self.__name += "_tar_max"
            else:
                self.__name += f"_tar_{target_mode}"
        if precision_target != 1.0:
            self.__name += f"_pt_{precision_target}"
        self.target_acceptance = target_acceptance
        self.target_mode = target_mode
        self.average_prediction = average_prediction
        self.marginal = marginal
        self.precision_target = precision_target

    def _additional_parameters(self, raw_data):
        biased_acceptance = compute_bias(raw_data["ytrain"], raw_data["strain"])

        if self.target_acceptance is None:
            if self.target_mode == self.MEAN:
                target_rate = 0.5 * (biased_acceptance[0] + biased_acceptance[1])
            elif self.target_mode == self.MIN:
                target_rate = min(biased_acceptance[0], biased_acceptance[1])
            elif self.target_mode == self.MAX:
                target_rate = max(biased_acceptance[0], biased_acceptance[1])
            else:
                acc_min = min(biased_acceptance[0], biased_acceptance[1])
                acc_max = max(biased_acceptance[0], biased_acceptance[1])
                target_rate = acc_min + self.target_mode * (acc_max - acc_min)
        else:
            target_rate = self.target_acceptance

        if self.marginal:
            p_s = prior_s(raw_data["strain"])
        else:
            p_s = [0.5] * 2

        return dict(
            lik="TunePrLikelihood",
            target_rate1=target_rate[0] if isinstance(target_rate, tuple) else target_rate,
            target_rate2=target_rate[1] if isinstance(target_rate, tuple) else target_rate,
            biased_acceptance1=biased_acceptance[0],
            biased_acceptance2=biased_acceptance[1],
            probs_from_flipped=False,
            average_prediction=self.average_prediction,
            p_s0=p_s[0],
            p_s1=p_s[1],
            p_ybary0_or_ybary1_s0=self.precision_target,
            p_ybary0_or_ybary1_s1=self.precision_target,
        )

    @property
    def name(self):
        """Getter for the algorithm name."""
        return self.__name


class GPyTEqOdds(GPyT):
    """GP algorithm which enforces equality of opportunity."""

    def __init__(
        self,
        average_prediction=False,
        tpr=None,
        marginal=False,
        tnr0=None,
        tnr1=None,
        tpr0=None,
        tpr1=None,
        **kwargs,
    ):
        """Init GP with eq. odds."""
        super().__init__(**kwargs)
        if self.s_as_input and average_prediction:
            self.__name = "{self.basename}_eq_odds_av_True"
            if marginal:
                self.__name += "_marg"
        else:
            self.__name = f"{self.basename}_eq_odds_in_{self.s_as_input}"

        self.odds: Optional[Dict[str, float]] = None
        if any(x is not None for x in [tnr0, tnr1, tpr0, tpr1]):  # if any of them is not `None`
            self.odds = {}
            for val, name, target in [
                (tnr0, "0tnr", "p_ybary0_s0"),
                (tnr1, "1tnr", "p_ybary0_s1"),
                (tpr0, "0tpr", "p_ybary1_s0"),
                (tpr1, "1tpr", "p_ybary1_s1"),
            ]:
                if val is not None:
                    self.odds[target] = val
                    self.__name += f"_{name}_{val}"  # add to name
                else:
                    self.odds[target] = 1.0  # default value
        elif tpr is not None:
            self.odds = dict(p_ybary0_s0=1.0, p_ybary0_s1=1.0, p_ybary1_s0=tpr, p_ybary1_s1=tpr)
            self.__name += f"_tpr_{tpr}"

        self.average_prediction = average_prediction
        self.marginal = marginal

    def _additional_parameters(self, raw_data):
        biased_acceptance = compute_bias(raw_data["ytrain"], raw_data["strain"])

        if self.marginal:
            p_s = prior_s(raw_data["strain"])
        else:
            p_s = [0.5] * 2

        return dict(
            lik="TuneTprLikelihood",
            p_ybary0_s0=1.0,
            p_ybary0_s1=1.0,
            p_ybary1_s0=1.0,
            p_ybary1_s1=1.0,
            biased_acceptance1=biased_acceptance[0],
            biased_acceptance2=biased_acceptance[1],
            average_prediction=self.average_prediction,
            p_s0=p_s[0],
            p_s1=p_s[1],
        )

    @implements(InAlgorithmAsync)
    async def run_async(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        (ytrain,), label_converter = _fix_labels([train.y.to_numpy()])
        raw_data = dict(
            xtrain=train.x.to_numpy(),
            strain=train.s.to_numpy(),
            ytrain=ytrain,
            xtest=test.x.to_numpy(),
            stest=test.s.to_numpy(),
            # ytest=ytest,
        )
        parameters = self._additional_parameters(raw_data)

        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            data_path = tmp_path / Path("data.npz")
            model_name = "local"  # f"run{self.counter}_s_as_input_{self.s_as_input}"
            flags = _flags(
                parameters,
                str(data_path),
                tmpdir,
                self.s_as_input,
                model_name,
                len(raw_data["ytrain"]),
            )

            if self.odds is None:
                # Split the training data into train and dev and save it to `data.npz`
                train_dev_data = split_train_dev(
                    raw_data["xtrain"], raw_data["ytrain"], raw_data["strain"]
                )
                np.savez(data_path, **train_dev_data)

                # First run
                await self._run_gpyt(flags)

                # Read the results from the numpy file 'predictions.npz'
                with (tmp_path / model_name / PRED_FNAME).open("rb") as file_obj:
                    output = np.load(file_obj)
                    prediction_on_train = output["pred_mean"]
                preds = (prediction_on_train > 0.5).astype(np.int32)
                odds = compute_odds(train_dev_data["ytest"], preds, train_dev_data["stest"])

                # Enforce equality of opportunity
                opportunity = min(odds["p_ybary1_s0"], odds["p_ybary1_s1"])
                odds["p_ybary1_s0"] = opportunity
                odds["p_ybary1_s1"] = opportunity
                flags.update({"epochs": 2 * flags["epochs"], **odds})
            else:
                flags.update(self.odds)

            # Save with real test data
            np.savez(data_path, **raw_data)

            # Second run
            await self._run_gpyt(flags)

            # Read the results from the numpy file 'predictions.npz'
            with (tmp_path / model_name / PRED_FNAME).open("rb") as file_obj:
                output = np.load(file_obj)
                pred_mean = output["pred_mean"]

        # Convert the result to the expected format
        predictions = label_converter((pred_mean > 0.5).astype(raw_data["ytrain"].dtype)[:, 0])
        return pd.DataFrame(predictions, columns=["preds"])

    @property
    def name(self):
        """Getter for algorithm name."""
        return self.__name


def prior_s(sensitive):
    """Compute the bias in the labels with respect to the sensitive attributes."""
    return (np.sum(sensitive == 0) / len(sensitive), np.sum(sensitive == 1) / len(sensitive))


def compute_bias(labels, sensitive):
    """Compute the bias in the labels with respect to the sensitive attributes."""
    rate_y1_s0 = np.sum(labels[sensitive == 0] == 1) / np.sum(sensitive == 0)
    rate_y1_s1 = np.sum(labels[sensitive == 1] == 1) / np.sum(sensitive == 1)
    return rate_y1_s0, rate_y1_s1


def compute_odds(labels, predictions, sensitive):
    """Compute the bias in the predictions with respect to the sensitive attr. and the labels."""
    return dict(
        p_ybary0_s0=np.mean(predictions[(labels == 0) & (sensitive == 0)] == 0),
        p_ybary1_s0=np.mean(predictions[(labels == 1) & (sensitive == 0)] == 1),
        p_ybary0_s1=np.mean(predictions[(labels == 0) & (sensitive == 1)] == 0),
        p_ybary1_s1=np.mean(predictions[(labels == 1) & (sensitive == 1)] == 1),
    )


def _fix_labels(labels):
    """Make sure that labels are either 0 or 1.

    Args"
        labels: the labels as a list of numpy arrays
        positive_class_val: the value that corresponds to a "positive" predictions

    Returns:
        the fixed labels and a function to convert the fixed labels back to the original format
    """
    label_values = list(np.unique(labels[0]))
    if label_values == [0, 1]:

        def _do_nothing(inp):
            return inp

        return labels, _do_nothing
    if label_values == [-1, 1]:

        def _converter(label):
            return (2 * label) - 1

        return [(y + 1) / 2 for y in labels], _converter
    raise ValueError("Labels have unknown structure")


def split_train_dev(inputs, labels, sensitive):
    """Split the given data into train and dev set with the proportion of labels being preserved."""
    n_total = inputs.shape[0]
    idx_s0_y0 = ((sensitive == 0) & (labels == 0)).nonzero()[0]
    idx_s0_y1 = ((sensitive == 0) & (labels == 1)).nonzero()[0]
    idx_s1_y0 = ((sensitive == 1) & (labels == 0)).nonzero()[0]
    idx_s1_y1 = ((sensitive == 1) & (labels == 1)).nonzero()[0]

    train_fraction: List[int] = []
    test_fraction: List[int] = []
    for idx in [idx_s0_y0, idx_s0_y1, idx_s1_y0, idx_s1_y1]:
        np.random.shuffle(idx)

        split_idx = int(len(idx) * 0.5) + 1  # make sure the train part is at least half
        train_fraction_a = idx[:split_idx]
        test_fraction_a = idx[split_idx:]
        train_fraction += list(train_fraction_a)
        test_fraction += list(test_fraction_a)
    xtrain, ytrain, strain = (
        inputs[train_fraction],
        labels[train_fraction],
        sensitive[train_fraction],
    )
    # ensure that the train set has exactly the same size as the given set
    # (otherwise inducing inputs has wrong shape)
    return dict(
        xtrain=np.concatenate((xtrain, xtrain))[:n_total],
        ytrain=np.concatenate((ytrain, ytrain))[:n_total],
        strain=np.concatenate((strain, strain))[:n_total],
        xtest=inputs[test_fraction],
        ytest=labels[test_fraction],
        stest=sensitive[test_fraction],
    )


def _flags(parameters, data_path, save_dir, s_as_input, model_name, num_train):
    batch_size = min(MAX_BATCH_SIZE, num_train)
    # epochs = _num_epochs(num_train)
    epochs = 70
    return {
        **dict(
            inf="Variational",
            data="sensitive_from_numpy",
            dataset_path=data_path,
            cov="RBFKernel",
            mean="ZeroMean",
            optimizer="Adam",
            lr=0.05,
            # lr=0.1,
            model_name=model_name,
            batch_size=batch_size,
            # epochs=min(MAX_EPOCHS, _num_epochs(num_train)),
            epochs=min(epochs, MAX_EPOCHS),
            eval_epochs=100000,  # we can unfortunately not run evaluations because of no y_test
            summary_steps=100000,
            chkpt_epochs=100000,
            save_dir=save_dir,  # "/home/ubuntu/out2/",
            plot="",
            logging_steps=1,
            gpus='0',
            preds_path=PRED_FNAME,  # save the predictions into `predictions.npz`
            num_samples=20,
            optimize_inducing=True,
            length_scale=1.2,
            sf=1.0,
            iso=False,
            num_samples_pred=2000,
            s_as_input=s_as_input,
            # num_inducing=MAX_NUM_INDUCING,
            num_inducing=_num_inducing(num_train),
            manual_seed=SEED,
            # metrics=(
            #     "binary_accuracy,pred_rate_y1_s0,pred_rate_y1_s1,base_rate_y1_s0,base_rate_y1_s1,"
            #     "pred_odds_yhaty1_s0,pred_odds_yhaty1_s1,pred_odds_yhaty0_s0,pred_odds_yhaty0_s1"
            # ),
        ),
        **parameters,
    }


def _num_inducing(num_train):
    """Adaptive number of inducing inputs.

    num_train == 4,000 => num_inducing == 1121
    num_train == 20,000 => num_inducing == 2507
    """
    return int(2500 / 141 * np.sqrt(num_train))


def _num_epochs(num_train):
    """Adaptive number of epochs.

    num_train == 4,000 => num_epochs == 125.7
    num_train == 20,000 => num_epochs == 84
    """
    return int(1000 / np.power(num_train, 1 / 4))



"""implementation of Agarwal model."""
from pathlib import Path
from typing import List, Set, Optional, Union, Dict

from ethicml.utility.data_structures import (
    PathTuple,
    TestPathTuple,
    FairnessType,
    ClassifierType,
)
from .shared import settings_for_svm_lr, flag_interface
from .in_algorithm import InAlgorithmAsync


VALID_FAIRNESS: Set[FairnessType] = {"DP", "EqOd"}
VALID_MODELS: Set[ClassifierType] = {"LR", "SVM"}


class Agarwal(InAlgorithmAsync):
    """Agarwal class."""

    def __init__(
        self,
        fairness: FairnessType = "DP",
        classifier: ClassifierType = "LR",
        eps: float = 0.1,
        iters: int = 50,
        C: Optional[float] = None,
        kernel: Optional[str] = None,
    ):
        """Init Agarwal."""
        if fairness not in VALID_FAIRNESS:
            raise ValueError("results: fairness must be one of %r." % VALID_FAIRNESS)
        if classifier not in VALID_MODELS:
            raise ValueError("results: classifier must be one of %r." % VALID_MODELS)
        super().__init__()
        chosen_c, chosen_kernel = settings_for_svm_lr(classifier, C, kernel)
        self.flags: Dict[str, Union[str, float, int]] = {
            "classifier": classifier,
            "fairness": fairness,
            "eps": eps,
            "iters": iters,
            "C": chosen_c,
            "kernel": chosen_kernel,
        }

    def _script_command(
        self, train_paths: PathTuple, test_paths: TestPathTuple, pred_path: Path
    ) -> (List[str]):
        args = flag_interface(train_paths, test_paths, pred_path, self.flags)
        return ["-m", "ethicml.implementations.agarwal"] + args

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return f"Agarwal, {self.flags['classifier']}, {self.flags['fairness']}"



"""Class to describe features of the Compas dataset."""
from typing_extensions import Literal

from ethicml.data.dataset import Dataset


class Compas(Dataset):
    """Compas (or ProPublica) dataset."""

    def __init__(
        self, split: Literal["Sex", "Race", "Race-Sex"] = "Sex", discrete_only: bool = False
    ):
        """Init Compas dataset."""
        super().__init__()
        self.split = split
        self.discrete_only = discrete_only
        self.features = [
            "sex",
            "age-num",
            "race",
            "juv-fel-count",
            "juv-misd-count",
            "juv-other-count",
            "priors-count",
            "two-year-recid",
            "age-cat_25 - 45",
            "age-cat_Greater than 45",
            "age-cat_Less than 25",
            "c-charge-degree_F",
            "c-charge-degree_M",
            "c-charge-desc_Abuse Without Great Harm",
            "c-charge-desc_Agg Abuse Elderlly/Disabled Adult",
            "c-charge-desc_Agg Assault W/int Com Fel Dome",
            "c-charge-desc_Agg Battery Grt/Bod/Harm",
            "c-charge-desc_Agg Fleeing and Eluding",
            "c-charge-desc_Agg Fleeing/Eluding High Speed",
            "c-charge-desc_Aggr Child Abuse-Torture,Punish",
            "c-charge-desc_Aggrav Battery w/Deadly Weapon",
            "c-charge-desc_Aggrav Child Abuse-Agg Battery",
            "c-charge-desc_Aggrav Child Abuse-Causes Harm",
            "c-charge-desc_Aggrav Stalking After Injunctn",
            "c-charge-desc_Aggravated Assault",
            "c-charge-desc_Aggravated Assault W/Dead Weap",
            "c-charge-desc_Aggravated Assault W/dead Weap",
            "c-charge-desc_Aggravated Assault W/o Firearm",
            "c-charge-desc_Aggravated Assault w/Firearm",
            "c-charge-desc_Aggravated Battery",
            "c-charge-desc_Aggravated Battery (Firearm)",
            "c-charge-desc_Aggravated Battery (Firearm/Actual Possession)",
            "c-charge-desc_Aggravated Battery / Pregnant",
            "c-charge-desc_Aggravated Battery On 65/Older",
            "c-charge-desc_Aide/Abet Prostitution Lewdness",
            "c-charge-desc_Aiding Escape",
            "c-charge-desc_Alcoholic Beverage Violation-FL",
            "c-charge-desc_Armed Trafficking in Cannabis",
            "c-charge-desc_Arson in the First Degree",
            "c-charge-desc_Assault",
            "c-charge-desc_Assault Law Enforcement Officer",
            "c-charge-desc_Att Burgl Conv Occp",
            "c-charge-desc_Att Burgl Struc/Conv Dwel/Occp",
            "c-charge-desc_Att Burgl Unoccupied Dwel",
            "c-charge-desc_Att Tamper w/Physical Evidence",
            "c-charge-desc_Attempt Armed Burglary Dwell",
            "c-charge-desc_Attempted Burg/Convey/Unocc",
            "c-charge-desc_Attempted Burg/struct/unocc",
            "c-charge-desc_Attempted Deliv Control Subst",
            "c-charge-desc_Attempted Robbery  No Weapon",
            "c-charge-desc_Attempted Robbery  Weapon",
            "c-charge-desc_Battery",
            "c-charge-desc_Battery Emergency Care Provide",
            "c-charge-desc_Battery On A Person Over 65",
            "c-charge-desc_Battery On Fire Fighter",
            "c-charge-desc_Battery On Parking Enfor Speci",
            "c-charge-desc_Battery Spouse Or Girlfriend",
            "c-charge-desc_Battery on Law Enforc Officer",
            "c-charge-desc_Battery on a Person Over 65",
            "c-charge-desc_Bribery Athletic Contests",
            "c-charge-desc_Burgl Dwel/Struct/Convey Armed",
            "c-charge-desc_Burglary Assault/Battery Armed",
            "c-charge-desc_Burglary Conveyance Armed",
            "c-charge-desc_Burglary Conveyance Assault/Bat",
            "c-charge-desc_Burglary Conveyance Occupied",
            "c-charge-desc_Burglary Conveyance Unoccup",
            "c-charge-desc_Burglary Dwelling Armed",
            "c-charge-desc_Burglary Dwelling Assault/Batt",
            "c-charge-desc_Burglary Dwelling Occupied",
            "c-charge-desc_Burglary Structure Assault/Batt",
            "c-charge-desc_Burglary Structure Occupied",
            "c-charge-desc_Burglary Structure Unoccup",
            "c-charge-desc_Burglary Unoccupied Dwelling",
            "c-charge-desc_Burglary With Assault/battery",
            "c-charge-desc_Carjacking w/o Deadly Weapon",
            "c-charge-desc_Carjacking with a Firearm",
            "c-charge-desc_Carry Open/Uncov Bev In Pub",
            "c-charge-desc_Carrying A Concealed Weapon",
            "c-charge-desc_Carrying Concealed Firearm",
            "c-charge-desc_Cash Item w/Intent to Defraud",
            "c-charge-desc_Child Abuse",
            "c-charge-desc_Computer Pornography",
            "c-charge-desc_Consp Traff Oxycodone  4g><14g",
            "c-charge-desc_Conspiracy Dealing Stolen Prop",
            "c-charge-desc_Consume Alcoholic Bev Pub",
            "c-charge-desc_Contradict Statement",
            "c-charge-desc_Contribute Delinquency Of A Minor",
            "c-charge-desc_Corrupt Public Servant",
            "c-charge-desc_Counterfeit Lic Plates/Sticker",
            "c-charge-desc_Crim Attempt/Solic/Consp",
            "c-charge-desc_Crim Use of Personal ID Info",
            "c-charge-desc_Crimin Mischief Damage $1000+",
            "c-charge-desc_Criminal Mischief",
            "c-charge-desc_Criminal Mischief Damage <$200",
            "c-charge-desc_Criminal Mischief>$200<$1000",
            "c-charge-desc_Crlty Twrd Child Urge Oth Act",
            "c-charge-desc_Cruelty Toward Child",
            "c-charge-desc_Cruelty to Animals",
            "c-charge-desc_Culpable Negligence",
            "c-charge-desc_D.U.I. Serious Bodily Injury",
            "c-charge-desc_DOC/Cause Public Danger",
            "c-charge-desc_DUI - Enhanced",
            "c-charge-desc_DUI - Property Damage/Personal Injury",
            "c-charge-desc_DUI Blood Alcohol Above 0.20",
            "c-charge-desc_DUI Level 0.15 Or Minor In Veh",
            "c-charge-desc_DUI Property Damage/Injury",
            "c-charge-desc_DUI- Enhanced",
            "c-charge-desc_DUI/Property Damage/Persnl Inj",
            "c-charge-desc_DWI w/Inj Susp Lic / Habit Off",
            "c-charge-desc_DWLS Canceled Disqul 1st Off",
            "c-charge-desc_DWLS Susp/Cancel Revoked",
            "c-charge-desc_Dealing in Stolen Property",
            "c-charge-desc_Defrauding Innkeeper",
            "c-charge-desc_Defrauding Innkeeper $300/More",
            "c-charge-desc_Del 3,4 Methylenedioxymethcath",
            "c-charge-desc_Del Cannabis At/Near Park",
            "c-charge-desc_Del Cannabis For Consideration",
            "c-charge-desc_Del of JWH-250 2-Methox 1-Pentyl",
            "c-charge-desc_Deliver 3,4 Methylenediox",
            "c-charge-desc_Deliver Alprazolam",
            "c-charge-desc_Deliver Cannabis",
            "c-charge-desc_Deliver Cannabis 1000FTSch",
            "c-charge-desc_Deliver Cocaine",
            "c-charge-desc_Deliver Cocaine 1000FT Church",
            "c-charge-desc_Deliver Cocaine 1000FT Park",
            "c-charge-desc_Deliver Cocaine 1000FT School",
            "c-charge-desc_Deliver Cocaine 1000FT Store",
            "c-charge-desc_Delivery Of Drug Paraphernalia",
            "c-charge-desc_Delivery of 5-Fluoro PB-22",
            "c-charge-desc_Delivery of Heroin",
            "c-charge-desc_Depriv LEO of Protect/Communic",
            "c-charge-desc_Disorderly Conduct",
            "c-charge-desc_Disorderly Intoxication",
            "c-charge-desc_Disrupting School Function",
            "c-charge-desc_Drivg While Lic Suspd/Revk/Can",
            "c-charge-desc_Driving License Suspended",
            "c-charge-desc_Driving Under The Influence",
            "c-charge-desc_Driving While License Revoked",
            "c-charge-desc_Escape",
            "c-charge-desc_Exhibition Weapon School Prop",
            "c-charge-desc_Expired DL More Than 6 Months",
            "c-charge-desc_Exposes Culpable Negligence",
            "c-charge-desc_Extradition/Defendants",
            "c-charge-desc_Fabricating Physical Evidence",
            "c-charge-desc_Fail Register Vehicle",
            "c-charge-desc_Fail Sex Offend Report Bylaw",
            "c-charge-desc_Fail To Obey Police Officer",
            "c-charge-desc_Fail To Redeliv Hire/Leas Prop",
            "c-charge-desc_Failure To Pay Taxi Cab Charge",
            "c-charge-desc_Failure To Return Hired Vehicle",
            "c-charge-desc_False 911 Call",
            "c-charge-desc_False Bomb Report",
            "c-charge-desc_False Imprisonment",
            "c-charge-desc_False Info LEO During Invest",
            "c-charge-desc_False Motor Veh Insurance Card",
            "c-charge-desc_False Name By Person Arrest",
            "c-charge-desc_False Ownership Info/Pawn Item",
            "c-charge-desc_Falsely Impersonating Officer",
            "c-charge-desc_Fel Drive License Perm Revoke",
            "c-charge-desc_Felon in Pos of Firearm or Amm",
            "c-charge-desc_Felony Batt(Great Bodily Harm)",
            "c-charge-desc_Felony Battery",
            "c-charge-desc_Felony Battery (Dom Strang)",
            "c-charge-desc_Felony Battery w/Prior Convict",
            "c-charge-desc_Felony Committing Prostitution",
            "c-charge-desc_Felony DUI (level 3)",
            "c-charge-desc_Felony DUI - Enhanced",
            "c-charge-desc_Felony Driving While Lic Suspd",
            "c-charge-desc_Felony Petit Theft",
            "c-charge-desc_Felony/Driving Under Influence",
            "c-charge-desc_Fighting/Baiting Animals",
            "c-charge-desc_Fleeing Or Attmp Eluding A Leo",
            "c-charge-desc_Fleeing or Eluding a LEO",
            "c-charge-desc_Forging Bank Bills/Promis Note",
            "c-charge-desc_Fraudulent Use of Credit Card",
            "c-charge-desc_Grand Theft (Motor Vehicle)",
            "c-charge-desc_Grand Theft Dwell Property",
            "c-charge-desc_Grand Theft Firearm",
            "c-charge-desc_Grand Theft in the 1st Degree",
            "c-charge-desc_Grand Theft in the 3rd Degree",
            "c-charge-desc_Grand Theft of a Fire Extinquisher",
            "c-charge-desc_Grand Theft of the 2nd Degree",
            "c-charge-desc_Grand Theft on 65 Yr or Older",
            "c-charge-desc_Harass Witness/Victm/Informnt",
            "c-charge-desc_Harm Public Servant Or Family",
            "c-charge-desc_Hiring with Intent to Defraud",
            "c-charge-desc_Imperson Public Officer or Emplyee",
            "c-charge-desc_Interfere W/Traf Cont Dev RR",
            "c-charge-desc_Interference with Custody",
            "c-charge-desc_Intoxicated/Safety Of Another",
            "c-charge-desc_Introduce Contraband Into Jail",
            "c-charge-desc_Issuing a Worthless Draft",
            "c-charge-desc_Kidnapping / Domestic Violence",
            "c-charge-desc_Lease For Purpose Trafficking",
            "c-charge-desc_Leave Acc/Attend Veh/More $50",
            "c-charge-desc_Leave Accd/Attend Veh/Less $50",
            "c-charge-desc_Leaving Acc/Unattended Veh",
            "c-charge-desc_Leaving the Scene of Accident",
            "c-charge-desc_Lewd Act Presence Child 16-",
            "c-charge-desc_Lewd or Lascivious Molestation",
            "c-charge-desc_Lewd/Lasc Battery Pers 12+/<16",
            "c-charge-desc_Lewd/Lasc Exhib Presence <16yr",
            "c-charge-desc_Lewd/Lasciv Molest Elder Persn",
            "c-charge-desc_Lewdness Violation",
            "c-charge-desc_License Suspended Revoked",
            "c-charge-desc_Littering",
            "c-charge-desc_Live on Earnings of Prostitute",
            "c-charge-desc_Lve/Scen/Acc/Veh/Prop/Damage",
            "c-charge-desc_Manage Busn W/O City Occup Lic",
            "c-charge-desc_Manslaughter W/Weapon/Firearm",
            "c-charge-desc_Manufacture Cannabis",
            "c-charge-desc_Misuse Of 911 Or E911 System",
            "c-charge-desc_Money Launder 100K or More Dols",
            "c-charge-desc_Murder In 2nd Degree W/firearm",
            "c-charge-desc_Murder in the First Degree",
            "c-charge-desc_Neglect Child / Bodily Harm",
            "c-charge-desc_Neglect Child / No Bodily Harm",
            "c-charge-desc_Neglect/Abuse Elderly Person",
            "c-charge-desc_Obstruct Fire Equipment",
            "c-charge-desc_Obstruct Officer W/Violence",
            "c-charge-desc_Obtain Control Substance By Fraud",
            "c-charge-desc_Offer Agree Secure For Lewd Act",
            "c-charge-desc_Offer Agree Secure/Lewd Act",
            "c-charge-desc_Offn Against Intellectual Prop",
            "c-charge-desc_Open Carrying Of Weapon",
            "c-charge-desc_Oper Motorcycle W/O Valid DL",
            "c-charge-desc_Operating W/O Valid License",
            "c-charge-desc_Opert With Susp DL 2nd Offens",
            "c-charge-desc_PL/Unlaw Use Credit Card",
            "c-charge-desc_Petit Theft",
            "c-charge-desc_Petit Theft $100- $300",
            "c-charge-desc_Pos Cannabis For Consideration",
            "c-charge-desc_Pos Cannabis W/Intent Sel/Del",
            "c-charge-desc_Pos Methylenedioxymethcath W/I/D/S",
            "c-charge-desc_Poss 3,4 MDMA (Ecstasy)",
            "c-charge-desc_Poss Alprazolam W/int Sell/Del",
            "c-charge-desc_Poss Anti-Shoplifting Device",
            "c-charge-desc_Poss Cntrft Contr Sub w/Intent",
            "c-charge-desc_Poss Cocaine/Intent To Del/Sel",
            "c-charge-desc_Poss Contr Subst W/o Prescript",
            "c-charge-desc_Poss Counterfeit Payment Inst",
            "c-charge-desc_Poss Drugs W/O A Prescription",
            "c-charge-desc_Poss F/Arm Delinq",
            "c-charge-desc_Poss Firearm W/Altered ID#",
            "c-charge-desc_Poss Meth/Diox/Meth/Amp (MDMA)",
            "c-charge-desc_Poss Of 1,4-Butanediol",
            "c-charge-desc_Poss Of Controlled Substance",
            "c-charge-desc_Poss Of RX Without RX",
            "c-charge-desc_Poss Oxycodone W/Int/Sell/Del",
            "c-charge-desc_Poss Pyrrolidinobutiophenone",
            "c-charge-desc_Poss Pyrrolidinovalerophenone",
            "c-charge-desc_Poss Pyrrolidinovalerophenone W/I/D/S",
            "c-charge-desc_Poss Similitude of Drivers Lic",
            "c-charge-desc_Poss Tetrahydrocannabinols",
            "c-charge-desc_Poss Unlaw Issue Driver Licenc",
            "c-charge-desc_Poss Unlaw Issue Id",
            "c-charge-desc_Poss Wep Conv Felon",
            "c-charge-desc_Poss of Cocaine W/I/D/S 1000FT Park",
            "c-charge-desc_Poss of Firearm by Convic Felo",
            "c-charge-desc_Poss of Methylethcathinone",
            "c-charge-desc_Poss/Sell/Del Cocaine 1000FT Sch",
            "c-charge-desc_Poss/Sell/Del/Man Amobarbital",
            "c-charge-desc_Poss/pur/sell/deliver Cocaine",
            "c-charge-desc_Poss3,4 Methylenedioxymethcath",
            "c-charge-desc_Posses/Disply Susp/Revk/Frd DL",
            "c-charge-desc_Possess Cannabis 1000FTSch",
            "c-charge-desc_Possess Cannabis/20 Grams Or Less",
            "c-charge-desc_Possess Controlled Substance",
            "c-charge-desc_Possess Countrfeit Credit Card",
            "c-charge-desc_Possess Drug Paraphernalia",
            "c-charge-desc_Possess Mot Veh W/Alt Vin #",
            "c-charge-desc_Possess Tobacco Product Under 18",
            "c-charge-desc_Possess Weapon On School Prop",
            "c-charge-desc_Possess w/I/Utter Forged Bills",
            "c-charge-desc_Possession Burglary Tools",
            "c-charge-desc_Possession Child Pornography",
            "c-charge-desc_Possession Firearm School Prop",
            "c-charge-desc_Possession Of 3,4Methylenediox",
            "c-charge-desc_Possession Of Alprazolam",
            "c-charge-desc_Possession Of Amphetamine",
            "c-charge-desc_Possession Of Anabolic Steroid",
            "c-charge-desc_Possession Of Buprenorphine",
            "c-charge-desc_Possession Of Carisoprodol",
            "c-charge-desc_Possession Of Clonazepam",
            "c-charge-desc_Possession Of Cocaine",
            "c-charge-desc_Possession Of Diazepam",
            "c-charge-desc_Possession Of Fentanyl",
            "c-charge-desc_Possession Of Heroin",
            "c-charge-desc_Possession Of Methamphetamine",
            "c-charge-desc_Possession Of Paraphernalia",
            "c-charge-desc_Possession Of Phentermine",
            "c-charge-desc_Possession of Alcohol Under 21",
            "c-charge-desc_Possession of Benzylpiperazine",
            "c-charge-desc_Possession of Butylone",
            "c-charge-desc_Possession of Cannabis",
            "c-charge-desc_Possession of Cocaine",
            "c-charge-desc_Possession of Codeine",
            "c-charge-desc_Possession of Ethylone",
            "c-charge-desc_Possession of Hydrocodone",
            "c-charge-desc_Possession of Hydromorphone",
            "c-charge-desc_Possession of LSD",
            "c-charge-desc_Possession of Methadone",
            "c-charge-desc_Possession of Morphine",
            "c-charge-desc_Possession of Oxycodone",
            "c-charge-desc_Possession of XLR11",
            "c-charge-desc_Principal In The First Degree",
            "c-charge-desc_Prostitution",
            "c-charge-desc_Prostitution/Lewd Act Assignation",
            "c-charge-desc_Prostitution/Lewdness/Assign",
            "c-charge-desc_Prowling/Loitering",
            "c-charge-desc_Purchase Cannabis",
            "c-charge-desc_Purchase/P/W/Int Cannabis",
            "c-charge-desc_Reckless Driving",
            "c-charge-desc_Refuse Submit Blood/Breath Test",
            "c-charge-desc_Refuse to Supply DNA Sample",
            "c-charge-desc_Resist Officer w/Violence",
            "c-charge-desc_Resist/Obstruct W/O Violence",
            "c-charge-desc_Retail Theft $300 1st Offense",
            "c-charge-desc_Retail Theft $300 2nd Offense",
            "c-charge-desc_Ride Tri-Rail Without Paying",
            "c-charge-desc_Robbery / No Weapon",
            "c-charge-desc_Robbery / Weapon",
            "c-charge-desc_Robbery Sudd Snatch No Weapon",
            "c-charge-desc_Robbery W/Deadly Weapon",
            "c-charge-desc_Robbery W/Firearm",
            "c-charge-desc_Sale/Del Cannabis At/Near Scho",
            "c-charge-desc_Sale/Del Counterfeit Cont Subs",
            "c-charge-desc_Sel/Pur/Mfr/Del Control Substa",
            "c-charge-desc_Sell or Offer for Sale Counterfeit Goods",
            "c-charge-desc_Sell/Man/Del Pos/w/int Heroin",
            "c-charge-desc_Sex Batt Faml/Cust Vict 12-17Y",
            "c-charge-desc_Sex Battery Deft 18+/Vict 11-",
            "c-charge-desc_Sex Offender Fail Comply W/Law",
            "c-charge-desc_Sexual Battery / Vict 12 Yrs +",
            "c-charge-desc_Sexual Performance by a Child",
            "c-charge-desc_Shoot In Occupied Dwell",
            "c-charge-desc_Shoot Into Vehicle",
            "c-charge-desc_Simulation of Legal Process",
            "c-charge-desc_Solic to Commit Battery",
            "c-charge-desc_Solicit Deliver Cocaine",
            "c-charge-desc_Solicit Purchase Cocaine",
            "c-charge-desc_Solicit To Deliver Cocaine",
            "c-charge-desc_Solicitation On Felony 3 Deg",
            "c-charge-desc_Soliciting For Prostitution",
            "c-charge-desc_Sound Articles Over 100",
            "c-charge-desc_Stalking",
            "c-charge-desc_Stalking (Aggravated)",
            "c-charge-desc_Strong Armed  Robbery",
            "c-charge-desc_Structuring Transactions",
            "c-charge-desc_Susp Drivers Lic 1st Offense",
            "c-charge-desc_Tamper With Victim",
            "c-charge-desc_Tamper With Witness",
            "c-charge-desc_Tamper With Witness/Victim/CI",
            "c-charge-desc_Tampering With Physical Evidence",
            "c-charge-desc_Tampering with a Victim",
            "c-charge-desc_Theft/To Deprive",
            "c-charge-desc_Threat Public Servant",
            "c-charge-desc_Throw Deadly Missile Into Veh",
            "c-charge-desc_Throw In Occupied Dwell",
            "c-charge-desc_Throw Missile Into Pub/Priv Dw",
            "c-charge-desc_Traff In Cocaine <400g>150 Kil",
            "c-charge-desc_Traffic Counterfeit Cred Cards",
            "c-charge-desc_Traffick Amphetamine 28g><200g",
            "c-charge-desc_Traffick Oxycodone     4g><14g",
            "c-charge-desc_Trans/Harm/Material to a Minor",
            "c-charge-desc_Trespass On School Grounds",
            "c-charge-desc_Trespass Other Struct/Conve",
            "c-charge-desc_Trespass Private Property",
            "c-charge-desc_Trespass Property w/Dang Weap",
            "c-charge-desc_Trespass Struct/Conveyance",
            "c-charge-desc_Trespass Structure w/Dang Weap",
            "c-charge-desc_Trespass Structure/Conveyance",
            "c-charge-desc_Trespassing/Construction Site",
            "c-charge-desc_Tresspass Struct/Conveyance",
            "c-charge-desc_Tresspass in Structure or Conveyance",
            "c-charge-desc_Unauth C/P/S Sounds>1000/Audio",
            "c-charge-desc_Unauth Poss ID Card or DL",
            "c-charge-desc_Unauthorized Interf w/Railroad",
            "c-charge-desc_Unl/Disturb Education/Instui",
            "c-charge-desc_Unlaw Lic Use/Disply Of Others",
            "c-charge-desc_Unlaw LicTag/Sticker Attach",
            "c-charge-desc_Unlaw Use False Name/Identity",
            "c-charge-desc_Unlawful Conveyance of Fuel",
            "c-charge-desc_Unlicensed Telemarketing",
            "c-charge-desc_Use Computer for Child Exploit",
            "c-charge-desc_Use Of 2 Way Device To Fac Fel",
            "c-charge-desc_Use Scanning Device to Defraud",
            "c-charge-desc_Use of Anti-Shoplifting Device",
            "c-charge-desc_Uttering Forged Bills",
            "c-charge-desc_Uttering Forged Credit Card",
            "c-charge-desc_Uttering Worthless Check +$150",
            "c-charge-desc_Uttering a Forged Instrument",
            "c-charge-desc_Video Voyeur-<24Y on Child >16",
            "c-charge-desc_Viol Injunct Domestic Violence",
            "c-charge-desc_Viol Injunction Protect Dom Vi",
            "c-charge-desc_Viol Pretrial Release Dom Viol",
            "c-charge-desc_Viol Prot Injunc Repeat Viol",
            "c-charge-desc_Violation License Restrictions",
            "c-charge-desc_Violation Of Boater Safety Id",
            "c-charge-desc_Violation of Injunction Order/Stalking/Cyberstalking",
            "c-charge-desc_Voyeurism",
            "c-charge-desc_arrest case no charge",
        ]

        self.continuous_features = [
            "age-num",
            "juv-fel-count",
            "juv-misd-count",
            "juv-other-count",
            "priors-count",
        ]

        if split == "Sex":
            self.sens_attrs = ["sex"]
            self.s_prefix = ["sex"]
            self.class_labels = ["two-year-recid"]
            self.class_label_prefix = ["two-year-recid"]
        elif split == "Race":
            self.sens_attrs = ["race"]
            self.s_prefix = ["race"]
            self.class_labels = ["two-year-recid"]
            self.class_label_prefix = ["two-year-recid"]
        elif split == "Race-Sex":
            self.sens_attrs = ["sex", "race"]
            self.s_prefix = ["race", "sex"]
            self.class_labels = ["two-year-recid"]
            self.class_label_prefix = ["two-year-recid"]
        else:
            raise NotImplementedError

    @property
    def name(self) -> str:
        """Getter for dataset name."""
        return f"Compas {self.split}"

    @property
    def filename(self) -> str:
        """Getter for file name."""
        return "compas-recidivism.csv"



"""Implementation for Louizos et al Variational Fair Autoencoder."""
# pylint: disable=arguments-differ

from typing import List, Optional, Any, Tuple
import torch
from torch import nn, Tensor

from ethicml.data.dataset import Dataset
from ethicml.implementations.vfae_modules.encoder import Encoder
from ethicml.implementations.vfae_modules.decoder import Decoder


class VFAENetwork(nn.Module):
    """Implements a generative model with two layers of stochastic variables.

    Where both are conditional, i.e.:

    p(x, z1, z2, y | s) = p(z2) p(y) p(z1 | z2, y) p(x | z1, s)

    with q(z1 | x, s) q(z2 | z1, y) q(y | z1) being the variational posteriors.
    """

    def __init__(
        self,
        dataset: Dataset,
        supervised: bool,
        input_size: int,
        latent_dims: int,
        z1_enc_size: List[int],
        z2_enc_size: List[int],
        z1_dec_size: List[int],
    ):
        """Init VFAE Network."""
        super(VFAENetwork, self).__init__()
        torch.manual_seed(888)

        self.supervised = supervised

        self.z1_encoder = Encoder(z1_enc_size, input_size + 1, latent_dims)
        if self.supervised:
            self.z2_encoder = Encoder(z2_enc_size, latent_dims + 1, latent_dims)
            self.z1_decoder = Encoder(z1_dec_size, latent_dims + 1, latent_dims)
        self.x_dec = Decoder(dataset)
        self.ypred = nn.Linear(latent_dims, 1)

    def encode_z1(self, x: Tensor, s: Tensor) -> Tuple[Tensor, Tensor]:
        """Encode Z1."""
        return self.z1_encoder(torch.cat((x, s), 1))

    def encode_z2(self, z1: Tensor, y: Tensor) -> Tuple[Tensor, Tensor]:
        """Encode Z2."""
        return self.z2_encoder(torch.cat((z1, y), 1))

    def decode_z1(self, z2: Tensor, y: Tensor) -> Tuple[Tensor, Tensor]:
        """Decode Z1."""
        return self.z1_decoder(torch.cat((z2, y), 1))

    @staticmethod
    def reparameterize(mean: Tensor, logvar: Tensor) -> Tensor:
        """Reparametrization trick - Leaving as a method to try and control reproducability."""
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return eps.mul(std).add_(mean)  # type: ignore[attr-defined]

    def forward(  # type: ignore[override]
        self, x: Tensor, s: Tensor, y: Tensor
    ) -> Tuple[
        Tuple[Tensor, Tensor, Tensor],
        Optional[Tuple[Tensor, Tensor, Tensor]],
        Optional[Tuple[Tensor, Tensor, Tensor]],
        Tensor,
        Optional[Tensor],
    ]:
        """Forward pass for network."""
        z1_mu, z1_logvar = self.encode_z1(x, s)
        # z1 = F.sigmoid(reparameterize(z1_mu, z1_logvar))
        z1 = self.reparameterize(z1_mu, z1_logvar)

        z2_triplet: Optional[Tuple[Any, Any, Any]]
        z1_d_triplet: Optional[Tuple[Any, Any, Any]]
        y_pred: Optional[torch.Tensor]
        if self.supervised:
            z2_mu, z2_logvar = self.encode_z2(z1, y)
            # z2 = F.sigmoid(reparameterize(z2_mu, z2_logvar))
            z2 = self.reparameterize(z2_mu, z2_logvar)

            z1_dec_mu, z1_dec_logvar = self.decode_z1(z2, y)
            # z1_dec = F.sigmoid(reparameterize(z1_dec_mu, z1_dec_logvar))
            z1_dec = self.reparameterize(z1_dec_mu, z1_dec_logvar)

            x_dec = self.x_dec(z1_dec, s)

            # y_pred_hl = self.yp_hl(z1)
            y_pred = torch.sigmoid(self.ypred(z1))

            z2_triplet = z2, z2_mu, z2_logvar
            z1_d_triplet = z1_dec, z1_dec_mu, z1_dec_logvar

        else:
            x_dec = self.x_dec(z1, s)
            z2_triplet = None
            z1_d_triplet = None
            y_pred = None

        z1_triplet = z1, z1_mu, z1_logvar

        return z1_triplet, z2_triplet, z1_d_triplet, x_dec, y_pred



"""Collection of functions that enable parallelism."""
import os
import asyncio
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Generic,
    List,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)
from collections import OrderedDict
from dataclasses import dataclass

import pandas as pd
from tqdm import tqdm

from ethicml.utility.data_structures import DataTuple, TrainTestPair, TestTuple
from ethicml.algorithms.inprocess import InAlgorithm, InAlgorithmAsync
from ethicml.algorithms.preprocess import PreAlgorithm, PreAlgorithmAsync

__all__ = ["arrange_in_parallel", "run_in_parallel"]

_AT = TypeVar("_AT", InAlgorithmAsync, PreAlgorithmAsync)  # async algorithm type
_ST = TypeVar("_ST", InAlgorithm, PreAlgorithm)  # sync algorithm type
_RT = TypeVar("_RT", pd.DataFrame, Tuple[DataTuple, TestTuple])  # return type

RunType = Callable[[DataTuple, TestTuple], Coroutine[Any, Any, _RT]]
BlockingRunType = Callable[[DataTuple, TestTuple], _RT]
DataSeq = Sequence[TrainTestPair]


@dataclass
class _Task(Generic[_RT]):
    algo_id: int
    data_id: int
    algo: Tuple[RunType[_RT], str]
    train_test_pair: TrainTestPair


async def arrange_in_parallel(
    algos: Sequence[Tuple[RunType[_RT], str]], data: DataSeq, max_parallel: int = 0,
) -> List[List[_RT]]:
    """Arrange the given algorithms to run (embarrassingly) parallel.

    Args:
        algos: list of tuples consisting of a `run_async` function of an algorithm and a name
        data: list of pairs of data tuples (train and test)
        max_parallel: how many processes can run in parallel at most. if zero (or negative), then
                      there is no maximum
    Returns:
        list of the results
    """
    assert len(algos) >= 1
    assert len(data) >= 1
    assert isinstance(data[0][0], DataTuple)
    assert isinstance(data[0][1], TestTuple)
    # ================================== create queue of tasks ====================================
    task_queue: asyncio.Queue[_Task[_RT]] = asyncio.Queue()
    # for each algorithm, first loop over all available datasets and then go on to the next algo
    for i, algo in enumerate(algos):
        for j, data_item in enumerate(data):
            task_queue.put_nowait(_Task(i, j, algo, data_item))

    num_tasks = len(algos) * len(data)
    pbar: tqdm = tqdm(total=num_tasks)

    # ===================================== create workers ========================================
    num_cpus = os.cpu_count()
    default_num_workers: int = num_cpus if num_cpus is not None else 1
    num_workers = max_parallel if max_parallel > 0 else default_num_workers
    result_dict: Dict[Tuple[int, int], _RT] = {}
    workers = [
        _eval_worker(worker_id, task_queue, result_dict, pbar) for worker_id in range(num_workers)
    ]
    # ====================================== run workers ==========================================
    await asyncio.gather(*workers)
    # confirm that the queue is empty
    assert task_queue.empty()
    pbar.close()  # very important! when we're not using "with", we have to close tqdm manually
    # turn dictionary into a list; the outer list is over the algos, the inner over the datasets
    # NOTE: if you want to change the return type, be warned that CrossValidator depends on it
    return [[result_dict[(i, j)] for j in range(len(data))] for i in range(len(algos))]


async def _eval_worker(
    worker_id: int,
    task_queue: "asyncio.Queue[_Task[_RT]]",
    result_dict: Dict[Tuple[int, int], _RT],
    pbar: tqdm,
) -> None:
    while not task_queue.empty():
        # get a work item out of the queue
        task = task_queue.get_nowait()
        (run_algo, algo_name) = task.algo
        train, test = task.train_test_pair
        # do some logging
        logging: OrderedDict[str, str] = OrderedDict()
        logging['model'] = algo_name
        logging['dataset'] = train.name if train.name is not None else ""
        logging['worker_id'] = str(worker_id)
        pbar.set_postfix(ordered_dict=logging)
        # do the work
        result: _RT = await run_algo(train, test)
        # put result into results dictionary
        result_dict[(task.algo_id, task.data_id)] = result
        pbar.update()
        # notify the queue that the work item has been processed
        task_queue.task_done()


InSeq = Sequence[InAlgorithm]
PreSeq = Sequence[PreAlgorithm]
InResult = List[List[pd.DataFrame]]
PreResult = List[List[Tuple[DataTuple, TestTuple]]]


@overload
async def run_in_parallel(algos: InSeq, data: DataSeq, max_parallel: int = 0) -> InResult:
    ...


@overload
async def run_in_parallel(algos: PreSeq, data: DataSeq, max_parallel: int = 0) -> PreResult:
    ...


async def run_in_parallel(
    algos: Union[InSeq, PreSeq], data: DataSeq, max_parallel: int = 0,
) -> Union[InResult, PreResult]:
    """Run the given algorithms (embarrassingly) parallel.

    Args:
        algos: list of algorithms
        data: list of pairs of data tuples (train and test)
        max_parallel: how many processes can run in parallel at most. if zero (or negative), then
                      there is no maximum
    Returns:
        list of the results
    """
    if not algos:
        return cast(List[List[pd.DataFrame]], [[]])
    if isinstance(algos[0], InAlgorithm):  # pylint: disable=no-else-return  # mypy needs this
        in_algos = cast(Sequence[InAlgorithm], algos)
        # Mypy complains in the next line because of https://github.com/python/mypy/issues/5374
        in_async_algos, async_idx, in_blocking_algos = _filter(in_algos, InAlgorithmAsync)  # type: ignore[misc]
        return await _generic_run_in_parallel(
            async_algos=[(algo.run_async, algo.name) for algo in in_async_algos],
            async_idx=async_idx,
            blocking_algos=[(algo.run, algo.name) for algo in in_blocking_algos],
            data=data,
            max_parallel=max_parallel,
        )
    elif isinstance(algos[0], PreAlgorithm):
        pre_algos = cast(Sequence[PreAlgorithm], algos)
        # Mypy complains in the next line because of https://github.com/python/mypy/issues/5374
        pre_async_algos, async_idx, pre_blocking_algos = _filter(pre_algos, PreAlgorithmAsync)  # type: ignore[misc]
        return await _generic_run_in_parallel(
            async_algos=[(algo.run_async, algo.name) for algo in pre_async_algos],
            async_idx=async_idx,
            blocking_algos=[(algo.run, algo.name) for algo in pre_blocking_algos],
            data=data,
            max_parallel=max_parallel,
        )


def _filter(algos: Sequence[_ST], algo_type: Type[_AT]) -> Tuple[List[_AT], List[int], List[_ST]]:
    # Filter out those algorithms that actually can run in their own process.
    # This is unfortunately really complicated because we have to keep track of the indices
    # in order to reassemble the returned list correctly.
    filtered_algos: List[_AT] = []
    filtered_idx: List[int] = []
    remaining_algos: List[_ST] = []
    for i, algo in enumerate(algos):
        if isinstance(algo, algo_type):
            filtered_algos.append(algo)
            filtered_idx.append(i)
        else:
            remaining_algos.append(algo)
    return filtered_algos, filtered_idx, remaining_algos


async def _generic_run_in_parallel(
    async_algos: Sequence[Tuple[RunType[_RT], str]],
    async_idx: List[int],
    blocking_algos: Sequence[Tuple[BlockingRunType[_RT], str]],
    data: DataSeq,
    max_parallel: int,
) -> List[List[_RT]]:
    """Generic version of `run_in_parallel` that allows us to do this with type safety."""
    # first start the asynchronous results
    if async_algos:
        async_coroutines = arrange_in_parallel(async_algos, data, max_parallel)

    # then get the blocking results
    # for each algorithm, first loop over all available datasets and then go on to the next algo
    blocking_results = [[run(train, test) for train, test in data] for run, name in blocking_algos]

    # then wait for the asynchronous results to come in
    if async_algos:
        async_results = await async_coroutines
    else:
        async_results = []

    # now reassemble everything in the right order
    results: List[List[_RT]] = []
    async_counter = 0
    blocking_counter = 0
    for i in range(len(async_algos) + len(blocking_algos)):
        if i in async_idx:
            results.append(async_results[async_counter])
            async_counter += 1
        else:
            results.append(blocking_results[blocking_counter])
            blocking_counter += 1
    return results



"""Pre-process algorithms take the training data and transform it."""
from .beutel import Beutel
from .pre_algorithm import PreAlgorithmAsync, PreAlgorithm
from .upsampler import Upsampler
from .vfae import VFAE
from .zemel import Zemel



"""Useful functions used in implementations."""
from pathlib import Path
from typing import Tuple, Union

import pandas as pd
import numpy as np
import tap

from ethicml.utility.data_structures import DataTuple, TestTuple, PathTuple, TestPathTuple


class AlgoArgs(tap.Tap):
    """ArgumentParser that already has arguments for the input data paths.

    This is a quick way to create a parser that can parse the filenames for the data. This class
    can be used by pre-algorithms to implement a commandline interface.
    """

    # paths to the files with the data
    train_x: str
    train_s: str
    train_y: str
    train_name: str
    test_x: str
    test_s: str
    test_name: str


class PreAlgoArgs(AlgoArgs):
    """ArgumentParser for pre-processing algorithms."""

    # paths to where the processed inputs should be stored
    new_train_x: str
    new_train_s: str
    new_train_y: str
    new_train_name: str
    new_test_x: str
    new_test_s: str
    new_test_name: str


class InAlgoArgs(AlgoArgs):
    """ArgumentParser that already has arguments for the paths needed for InAlgorithms."""

    # path to where the predictions should be stored
    pred_path: str


def save_predictions(predictions: Union[np.ndarray, pd.DataFrame], args: InAlgoArgs):
    """Save the data to the file that was specified in the commandline arguments."""
    if not isinstance(predictions, pd.DataFrame):
        df = pd.DataFrame(predictions, columns=["pred"])
    else:
        df = predictions
    pred_path = Path(args.pred_path)
    df.to_feather(pred_path)


def load_data_from_flags(flags: AlgoArgs) -> Tuple[DataTuple, TestTuple]:
    """Load data from the paths specified in the flags."""
    train_paths = PathTuple(
        x=Path(flags.train_x), s=Path(flags.train_s), y=Path(flags.train_y), name=flags.train_name
    )
    test_paths = TestPathTuple(x=Path(flags.test_x), s=Path(flags.test_s), name=flags.test_name)
    return train_paths.load_from_feather(), test_paths.load_from_feather()


def save_transformations(transforms: Tuple[DataTuple, TestTuple], args: PreAlgoArgs):
    """Save the data to the file that was specified in the commandline arguments."""
    assert isinstance(transforms[0], DataTuple)
    assert isinstance(transforms[1], TestTuple)

    train, test = transforms

    def _save(data: pd.DataFrame, path: str) -> None:
        # write the file
        data.columns = data.columns.astype(str)
        data.to_feather(Path(path))

    _save(train.x, args.new_train_x)
    _save(train.s, args.new_train_s)
    _save(train.y, args.new_train_y)
    _save(test.x, args.new_test_x)
    _save(test.s, args.new_test_s)



"""Anti-spurious."""

import pandas as pd

from ethicml.common import implements
from ethicml.utility.data_structures import DataTuple
from .metric import Metric


class AS(Metric):
    r"""Anti-spurious metric.

    Computes :math:`P(\hat{y}=y|y\neq s)`.
    """

    @implements(Metric)
    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        preds = prediction.to_numpy()
        sens = actual.s.to_numpy()
        labels = actual.y.to_numpy()
        s_uneq_y = sens != labels
        return (preds[s_uneq_y] == labels[s_uneq_y]).mean()

    @property
    def name(self) -> str:
        """Getter for the metric name."""
        return "anti_spurious"



import pytest

import torch
from torch.utils.data import DataLoader, random_split

from ethicml.vision import LdColorizer


@pytest.mark.parametrize("transform", [LdColorizer])
def test_label_dependent_transforms(transform):
    """test label dependent transforms"""

    data = torch.rand((9, 3, 4, 4))
    labels = torch.randint(low=0, high=10, size=(9,))

    colorizer = LdColorizer(
        min_val=0.0,
        max_val=1.0,
        scale=0.02,
        binarize=False,
        background=False,
        black=False,
        seed=47,
        greyscale=False,
    )

    #   TODO: test with meaningful assertions
    colorizer(data, labels)



"""
This file is automatically imported by pytest (no need to import it) and defines shared fixtures
"""
from pathlib import Path
import shutil

import pytest

from ethicml.utility import DataTuple, TrainTestPair
from ethicml.data import load_data, Toy
from ethicml.preprocessing import train_test_split


@pytest.fixture(scope="session")
def toy_train_test() -> TrainTestPair:
    """By making this a fixture, pytest can cache the result"""
    data: DataTuple = load_data(Toy())
    train: DataTuple
    test: DataTuple
    train, test = train_test_split(data)
    return TrainTestPair(train, test.remove_y())


@pytest.fixture(scope="module")
def plot_cleanup():
    """
    Clean up after the tests by removing the `plots` and `results` directories
    """
    yield None
    print("remove generated directory")
    plt_dir = Path(".") / "plots"
    res_dir = Path(".") / "results"
    if plt_dir.exists():
        shutil.rmtree(plt_dir)
    if res_dir.exists():
        shutil.rmtree(res_dir)


@pytest.fixture(scope="function")
def results_cleanup():
    """
    Clean up after the tests by removing the `results` directory
    """
    yield None
    print("remove generated directory")
    res_dir = Path(".") / "results"
    if res_dir.exists():
        shutil.rmtree(res_dir)



"""
Test that an algorithm can run against some data
"""
import os
from pathlib import Path
from typing import Tuple, List
import numpy as np  # pylint: disable=unused-import  # import needed for mypy
import pandas as pd
import pytest

from ethicml.algorithms import run_blocking
from ethicml.algorithms.inprocess import LR, SVM, Majority, InAlgorithm
from ethicml.algorithms.postprocess import PostAlgorithm
from ethicml.algorithms.preprocess import PreAlgorithm, Upsampler
from ethicml.metrics import Accuracy, CV, TPR, Metric
from ethicml.utility import DataTuple, TrainTestPair
from ethicml.data import load_data, Toy, Adult, Dataset
from ethicml.evaluators import (
    run_in_parallel,
    evaluate_models,
    MetricNotApplicable,
    load_results,
    evaluate_models_async,
)
from ethicml.preprocessing import train_test_split


def get_train_test() -> Tuple[DataTuple, DataTuple]:
    """Helper function for other tests which loads the Toy dataset and splits it into train-test"""
    data: DataTuple = load_data(Toy())
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    return train_test


def count_true(mask: "np.ndarray[np.bool_]") -> int:
    """Count the number of elements that are True"""
    return mask.nonzero()[0].shape[0]


def test_can_load_test_data():
    """test whether we can load test data"""
    train, test = get_train_test()
    assert train is not None
    assert test is not None


def test_run_parallel():
    """test run parallel"""
    data0 = get_train_test()
    data1 = get_train_test()
    result = run_blocking(
        run_in_parallel(
            [LR(), SVM(), Majority()],
            [TrainTestPair(*data0), TrainTestPair(*data1)],
            max_parallel=2,
        )
    )
    # LR
    assert count_true(result[0][0].values == 1) == 211
    assert count_true(result[0][0].values == -1) == 189
    assert count_true(result[0][1].values == 1) == 211
    assert count_true(result[0][1].values == -1) == 189
    # SVM
    assert count_true(result[1][0].values == 1) == 201
    assert count_true(result[1][0].values == -1) == 199
    assert count_true(result[1][1].values == 1) == 201
    assert count_true(result[1][1].values == -1) == 199
    # Majority
    assert count_true(result[2][0].values == 1) == 0
    assert count_true(result[2][0].values == -1) == 400
    assert count_true(result[2][1].values == 1) == 0
    assert count_true(result[2][1].values == -1) == 400


@pytest.mark.usefixtures("results_cleanup")
def test_empty_evaluate():
    """test empty evaluate"""
    empty_result = evaluate_models([Toy()], repeats=3, delete_prev=True)
    expected_result = pd.DataFrame([], columns=["dataset", "transform", "model", "split_id"])
    expected_result = expected_result.set_index(["dataset", "transform", "model", "split_id"])
    pd.testing.assert_frame_equal(empty_result.data, expected_result)


@pytest.mark.usefixtures("results_cleanup")
def test_run_alg_suite():
    """test run alg suite"""
    dataset = Adult("Race")
    dataset.sens_attrs = ["race_White"]
    datasets: List[Dataset] = [dataset, Toy()]
    preprocess_models: List[PreAlgorithm] = [Upsampler()]
    inprocess_models: List[InAlgorithm] = [LR(), SVM(kernel='linear')]
    postprocess_models: List[PostAlgorithm] = []
    metrics: List[Metric] = [Accuracy(), CV()]
    per_sens_metrics: List[Metric] = [Accuracy(), TPR()]
    parallel_results = run_blocking(
        evaluate_models_async(
            datasets,
            preprocess_models,
            inprocess_models,
            postprocess_models,
            metrics,
            per_sens_metrics,
            repeats=1,
            test_mode=True,
            topic="pytest",
        )
    )
    results = evaluate_models(
        datasets,
        preprocess_models,
        inprocess_models,
        postprocess_models,
        metrics,
        per_sens_metrics,
        repeats=1,
        test_mode=True,
        delete_prev=True,
        topic="pytest",
    )
    pd.testing.assert_frame_equal(parallel_results.data, results.data, check_like=True)

    files = os.listdir(Path(".") / "results")
    file_names = [
        'pytest_Adult Race_Upsample uniform.csv',
        'pytest_Adult Race_no_transform.csv',
        'pytest_Toy_Upsample uniform.csv',
        'pytest_Toy_no_transform.csv',
    ]
    assert len(files) == 4
    assert sorted(files) == file_names

    for file in file_names:
        written_file = pd.read_csv(Path(f"./results/{file}"))
        assert (written_file['seed'][0], written_file['seed'][1]) == (0, 0)
        assert written_file.shape == (2, 15)

    reloaded = load_results("Adult Race", "Upsample uniform", "pytest")
    assert reloaded is not None
    read = pd.read_csv(Path(".") / "results" / 'pytest_Adult Race_Upsample uniform.csv')
    read = read.set_index(["dataset", "transform", "model", "split_id"])
    pd.testing.assert_frame_equal(reloaded.data, read)


@pytest.mark.usefixtures("results_cleanup")
def test_run_alg_suite_wrong_metrics():
    """test run alg suite wrong metrics"""
    datasets: List[Dataset] = [Toy(), Adult()]
    preprocess_models: List[PreAlgorithm] = [Upsampler()]
    inprocess_models: List[InAlgorithm] = [SVM(kernel='linear'), LR()]
    postprocess_models: List[PostAlgorithm] = []
    metrics: List[Metric] = [Accuracy(), CV()]
    per_sens_metrics: List[Metric] = [Accuracy(), TPR(), CV()]
    with pytest.raises(MetricNotApplicable):
        evaluate_models(
            datasets,
            preprocess_models,
            inprocess_models,
            postprocess_models,
            metrics,
            per_sens_metrics,
            repeats=1,
            test_mode=True,
        )



"""File For feature binning."""
from itertools import groupby
from typing import List

import pandas as pd
from ethicml.utility import DataTuple

__all__ = ['bin_cont_feats']


def bin_cont_feats(data: DataTuple) -> DataTuple:
    """Bin the continuous fetures.

    Given a datatuple, bin the columns that have ordinal features
    and return as afresh new DataTuple.
    """
    groups: List[List[str]] = [
        list(group) for _, group in groupby(data.x.columns, lambda x: x.split("_")[0])
    ]

    copy: pd.DataFrame = data.x.copy()

    for group in groups:
        # if there is only one element in the group, then it corresponds to a continuous feature
        if len(group) == 1 and data.x[group[0]].nunique() > 2:
            copy[group] = pd.cut(data.x[group].to_numpy()[:, 0], 5)
            copy = pd.concat([copy, pd.get_dummies(copy[group])], axis="columns")
            copy = copy.drop(group, axis="columns")

    return data.replace(x=copy)



"""Abstract Base Class of all algorithms in the framework."""
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List
from abc import abstractmethod

import pandas as pd

from ethicml.common import implements
from ethicml.algorithms.algorithm_base import Algorithm, AlgorithmAsync, run_blocking
from ethicml.utility.data_structures import (
    DataTuple,
    TestTuple,
    PathTuple,
    TestPathTuple,
    write_as_feather,
    load_feather,
)


class InAlgorithm(Algorithm):
    """Abstract Base Class for algorithms that run in the middle of the pipeline."""

    @abstractmethod
    def run(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        """Run Algorithm on the given data.

        Args:
            train: training data
            test: test data
        """

    def run_test(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        """Run with reduced training set so that it finishes quicker."""
        train_testing = train.get_subset()
        return self.run(train_testing, test)


class InAlgorithmAsync(InAlgorithm, AlgorithmAsync):
    """In-Algorithm that can be run blocking and asynchronously."""

    @implements(InAlgorithm)
    def run(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        """Run this asynchronous Algorithm as blocking on the given data.

        Args:
            train: training data
            test: test data
        """
        return run_blocking(self.run_async(train, test))

    async def run_async(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        """Run Algorithm on the given data asynchronously.

        Args:
            train: training data
            test: test data
        """
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            train_paths, test_paths = write_as_feather(train, test, tmp_path)
            pred_path = tmp_path / "predictions.feather"
            cmd = self._script_command(train_paths, test_paths, pred_path)
            await self._call_script(cmd)  # wait for scrip to run
            return load_feather(pred_path)

    @abstractmethod
    def _script_command(
        self, train_paths: PathTuple, test_paths: TestPathTuple, pred_path: Path
    ) -> (List[str]):
        """The command that will run the script."""



"""This module contains kind of useful things that don't really belong anywhere else (just yet)."""

from .activation import Activation
from .heaviside import Heaviside
from .data_structures import (
    ActivationType,
    DataTuple,
    FairnessType,
    PathTuple,
    Results,
    TestPathTuple,
    TestTuple,
    TrainTestPair,
    concat_dt,
    concat_tt,
    load_feather,
)



"""In-process algorithms take training data and make predictions."""
__all__ = [
    "Agarwal",
    "InAlgorithm",
    "InAlgorithmAsync",
    "InstalledModel",
    "GPyT",
    "GPyTDemPar",
    "GPyTEqOdds",
    "Kamishima",
    "Kamiran",
    "LR",
    "LRCV",
    "LRProb",
    "SVM",
    "Majority",
    "MLP",
    "Corels",
    "ZafarAccuracy",
    "ZafarBaseline",
    "ZafarEqOdds",
    "ZafarEqOpp",
    "ZafarFairness",
]
from .agarwal_reductions import Agarwal
from .in_algorithm import InAlgorithm, InAlgorithmAsync
from .installed_model import InstalledModel
from .fair_gpyt import GPyT, GPyTDemPar, GPyTEqOdds
from .kamishima import Kamishima
from .kamiran import Kamiran
from .logistic_regression import LR, LRCV, LRProb
from .svm import SVM
from .majority import Majority
from .mlp import MLP
from .manual import Corels
from .zafar import ZafarAccuracy, ZafarBaseline, ZafarEqOdds, ZafarEqOpp, ZafarFairness



"""Zemel's Learned Fair Representations."""
from typing import Dict, Union, List

from ethicml.utility.data_structures import PathTuple, TestPathTuple
from .pre_algorithm import PreAlgorithmAsync
from .interface import flag_interface


class Zemel(PreAlgorithmAsync):
    """AIF360 implementation of Zemel's LFR."""

    def __init__(
        self,
        threshold: float = 0.5,
        clusters: int = 2,
        Ax: float = 0.01,
        Ay: float = 0.1,
        Az: float = 0.5,
        max_iter: int = 5000,
        maxfun: int = 5000,
        epsilon: float = 1e-5,
    ) -> None:
        """Init Zemel."""
        super().__init__()
        self.flags: Dict[str, Union[int, float]] = {
            "clusters": clusters,
            "Ax": Ax,
            "Ay": Ay,
            "Az": Az,
            "max_iter": max_iter,
            "maxfun": maxfun,
            "epsilon": epsilon,
            "threshold": threshold,
        }

    def _script_command(
        self,
        train_paths: PathTuple,
        test_paths: TestPathTuple,
        new_train_paths: PathTuple,
        new_test_paths: TestPathTuple,
    ) -> List[str]:
        args = flag_interface(train_paths, test_paths, new_train_paths, new_test_paths, self.flags)
        return ["-m", "ethicml.implementations.zemel"] + args

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return "Zemel"



"""Abstract Base Class of all post-processing algorithms in the framework."""

from abc import abstractmethod
import pandas as pd

from ethicml.utility.data_structures import DataTuple, TestTuple
from ..algorithm_base import Algorithm


class PostAlgorithm(Algorithm):
    """Abstract Base Class for all algorithms that do post-processing."""

    @abstractmethod
    def run(
        self,
        train_predictions: pd.DataFrame,
        train: DataTuple,
        test_predictions: pd.DataFrame,
        test: TestTuple,
    ) -> pd.DataFrame:
        """Make predictions fair.

        Args:
            train_predictions: predictions on a training set
            train: the training set with the correct labels and sensitive attributes
            test_predictions: predictions on the test set
            test: the test set with the sensitive attributes
        Return:
            post-processed predictions on the test set
        """



"""Helper functions for working with DataFrames (and Series)."""
import pandas as pd


def shuffle_df(df: pd.DataFrame, random_state: int) -> pd.DataFrame:
    """Shuffle a given dataframe."""
    return df.sample(frac=1.0, random_state=random_state).reset_index(drop=True)



"""Wrapper around Sci-Kit Learn Logistic Regression."""
from typing import Optional

import pandas as pd
from sklearn.linear_model import LogisticRegression, LogisticRegressionCV
from sklearn.model_selection import KFold

from ethicml.common import implements
from ethicml.utility.data_structures import DataTuple, TestTuple
from .in_algorithm import InAlgorithm


class LR(InAlgorithm):
    """Logistic regression with hard predictions."""

    def __init__(self, C: Optional[float] = None):
        """Init LR."""
        super().__init__()
        self.C = LogisticRegression().C if C is None else C

    @implements(InAlgorithm)
    def run(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        clf = LogisticRegression(solver="liblinear", random_state=888, C=self.C, multi_class='auto')
        clf.fit(train.x, train.y.to_numpy().ravel())
        return pd.DataFrame(clf.predict(test.x), columns=["preds"])

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return f"Logistic Regression, C={self.C}"


class LRProb(InAlgorithm):
    """Logistic regression with soft output."""

    def __init__(self, C: Optional[int] = None):
        """Init LRProb."""
        super().__init__()
        self.C = LogisticRegression().C if C is None else C

    @implements(InAlgorithm)
    def run(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        clf = LogisticRegression(solver="liblinear", random_state=888, C=self.C, multi_class='auto')
        clf.fit(train.x, train.y.to_numpy().ravel())
        return pd.DataFrame(clf.predict_proba(test.x)[:, 1], columns=["preds"])

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return f"Logistic Regression Prob, C={self.C}"


class LRCV(InAlgorithm):
    """Kind of a cheap hack for now, but gives a proper cross-valudeted LR."""

    @implements(InAlgorithm)
    def run(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        folder = KFold(n_splits=3, random_state=888, shuffle=False)
        clf = LogisticRegressionCV(
            cv=folder, n_jobs=-1, random_state=888, solver="liblinear", multi_class='auto'
        )
        clf.fit(train.x, train.y.to_numpy().ravel())
        return pd.DataFrame(clf.predict(test.x), columns=["preds"])

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return "LRCV"



"""Class to describe features of the German dataset."""

from .dataset import Dataset


class German(Dataset):
    """German credit dataset."""

    def __init__(self, split: str = "Sex", discrete_only: bool = False):
        """Init German dataset."""
        super().__init__()
        self.split = split
        self.discrete_only = discrete_only
        self.features = [
            "month",
            "credit-amount",
            "investment-as-income-percentage",
            "sex",
            "residence-since",
            "age",
            "number-of-credits",
            "people-liable-for",
            "credit-label",
            "status_A11",
            "status_A12",
            "status_A13",
            "status_A14",
            "credit-history_A30",
            "credit-history_A31",
            "credit-history_A32",
            "credit-history_A33",
            "credit-history_A34",
            "purpose_A40",
            "purpose_A41",
            "purpose_A410",
            "purpose_A42",
            "purpose_A43",
            "purpose_A44",
            "purpose_A45",
            "purpose_A46",
            "purpose_A48",
            "purpose_A49",
            "savings_A61",
            "savings_A62",
            "savings_A63",
            "savings_A64",
            "savings_A65",
            "employment_A71",
            "employment_A72",
            "employment_A73",
            "employment_A74",
            "employment_A75",
            "other-debtors_A101",
            "other-debtors_A102",
            "other-debtors_A103",
            "property_A121",
            "property_A122",
            "property_A123",
            "property_A124",
            "installment-plans_A141",
            "installment-plans_A142",
            "installment-plans_A143",
            "housing_A151",
            "housing_A152",
            "housing_A153",
            "skill-level_A171",
            "skill-level_A172",
            "skill-level_A173",
            "skill-level_A174",
            "telephone_A191",
            "telephone_A192",
            "foreign-worker_A201",
            "foreign-worker_A202",
        ]

        self.continuous_features = [
            "month",
            "credit-amount",
            "investment-as-income-percentage",
            "residence-since",
            "age",
            "number-of-credits",
            "people-liable-for",
        ]

        if split == "Sex":
            self.sens_attrs = ["sex"]
            self.s_prefix = ["sex"]
            self.class_labels = ["credit-label"]
            self.class_label_prefix = ["credit-label"]
        else:
            raise NotImplementedError

    @property
    def name(self) -> str:
        """Getter for dataset name."""
        return f"German {self.split}"

    @property
    def filename(self) -> str:
        """Getter for filename."""
        return "german.csv"



"""Methods that are shared among the inprocess algorithms."""
from pathlib import Path
from typing import List, Tuple, Optional, Any, Dict
from dataclasses import asdict

from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC

from ethicml.utility.data_structures import PathTuple, TestPathTuple


def flag_interface(
    train_paths: PathTuple, test_paths: TestPathTuple, pred_path: Path, flags: Dict[str, Any],
) -> List[str]:
    """Generate the commandline arguments that are expected."""
    flags_list: List[str] = []

    # paths to training and test data
    path_tuples = [train_paths, test_paths]
    prefixes = ["--train_", "--test_"]
    for path_tuple, prefix in zip(path_tuples, prefixes):
        for key, path in asdict(path_tuple).items():
            flags_list += [f"{prefix}{key}", str(path)]

    # paths to output files
    flags_list += ["--pred_path", str(pred_path)]

    # model parameters
    for key, values in flags.items():
        flags_list.append(f"--{key}")
        if isinstance(values, (list, tuple)):
            flags_list += [str(value) for value in values]
        else:
            flags_list.append(str(values))
    return flags_list


def settings_for_svm_lr(
    classifier: str, C: Optional[float], kernel: Optional[str]
) -> Tuple[float, str]:
    """If necessary get the default settings for the C and kernel parameter of SVM and LR."""
    if C is None:
        if classifier == "LR":
            C = LogisticRegression().C
        elif classifier == "SVM":
            C = SVC().C
        else:
            raise NotImplementedError(f"Unsupported classifier \"{classifier}\".")

    if kernel is None:
        if classifier == "SVM":
            kernel = SVC().kernel
        else:
            kernel = ""

    return C, kernel



"""Kamiran and Calders 2012."""
from typing import Optional, Tuple

import pandas as pd
import numpy as np
from sklearn.linear_model import LogisticRegression

from ethicml.common import implements
from ethicml.utility.data_structures import DataTuple, TestTuple, ClassifierType
from .shared import settings_for_svm_lr
from .in_algorithm import InAlgorithm
from .svm import select_svm

VALID_MODELS = {"LR", "SVM"}


class Kamiran(InAlgorithm):
    """Kamiran and Calders 2012."""

    def __init__(
        self,
        classifier: ClassifierType = "LR",
        C: Optional[float] = None,
        kernel: Optional[str] = None,
    ):
        """Init Kamiran."""
        super().__init__()
        if classifier not in VALID_MODELS:
            raise ValueError(f"results: classifier must be one of {VALID_MODELS!r}.")
        self.classifier = classifier
        self.C, self.kernel = settings_for_svm_lr(classifier, C, kernel)

    @implements(InAlgorithm)
    def run(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        return train_and_predict(
            train, test, classifier=self.classifier, C=self.C, kernel=self.kernel
        )

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return f"Kamiran & Calders {self.classifier}"


def _obtain_conditionings(
    dataset: DataTuple,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Obtain the necessary conditioning boolean vectors to compute instance level weights."""
    y_col = dataset.y.columns[0]
    y_pos = dataset.y[y_col].max()
    y_neg = dataset.y[y_col].min()
    s_col = dataset.s.columns[0]
    s_pos = dataset.s[s_col].max()
    s_neg = dataset.s[s_col].min()

    # combination of label and privileged/unpriv. groups
    cond_p_fav = dataset.x.loc[(dataset.y[y_col] == y_pos) & (dataset.s[s_col] == s_pos)]
    cond_p_unfav = dataset.x.loc[(dataset.y[y_col] == y_neg) & (dataset.s[s_col] == s_pos)]
    cond_up_fav = dataset.x.loc[(dataset.y[y_col] == y_pos) & (dataset.s[s_col] == s_neg)]
    cond_up_unfav = dataset.x.loc[(dataset.y[y_col] == y_neg) & (dataset.s[s_col] == s_neg)]

    return cond_p_fav, cond_p_unfav, cond_up_fav, cond_up_unfav


def compute_weights(train: DataTuple) -> pd.DataFrame:
    """Compute weights for all samples."""
    np.random.seed(888)
    (cond_p_fav, cond_p_unfav, cond_up_fav, cond_up_unfav) = _obtain_conditionings(train)

    y_col = train.y.columns[0]
    y_pos = train.y[y_col].max()
    y_neg = train.y[y_col].min()
    s_col = train.s.columns[0]
    s_pos = train.s[s_col].max()
    s_neg = train.s[s_col].min()

    num_samples = train.x.shape[0]
    n_p = train.s.loc[train.s[s_col] == s_pos].shape[0]
    n_up = train.s.loc[train.s[s_col] == s_neg].shape[0]
    n_fav = train.y.loc[train.y[y_col] == y_pos].shape[0]
    n_unfav = train.y.loc[train.y[y_col] == y_neg].shape[0]

    n_p_fav = cond_p_fav.shape[0]
    n_p_unfav = cond_p_unfav.shape[0]
    n_up_fav = cond_up_fav.shape[0]
    n_up_unfav = cond_up_unfav.shape[0]

    w_p_fav = n_fav * n_p / (num_samples * n_p_fav) if n_p_fav != 0 else float('NaN')
    w_p_unfav = n_unfav * n_p / (num_samples * n_p_unfav) if n_p_unfav != 0 else float('NaN')
    w_up_fav = n_fav * n_up / (num_samples * n_up_fav) if n_up_fav != 0 else float('NaN')
    w_up_unfav = n_unfav * n_up / (num_samples * n_up_unfav) if n_up_unfav != 0 else float('NaN')

    train_instance_weights = pd.DataFrame(np.ones(train.x.shape[0]), columns=["instance weights"])

    train_instance_weights.iloc[cond_p_fav.index] *= w_p_fav
    train_instance_weights.iloc[cond_p_unfav.index] *= w_p_unfav
    train_instance_weights.iloc[cond_up_fav.index] *= w_up_fav
    train_instance_weights.iloc[cond_up_unfav.index] *= w_up_unfav

    return train_instance_weights


def train_and_predict(
    train: DataTuple, test: TestTuple, classifier: ClassifierType, C: float, kernel: str
) -> pd.DataFrame:
    """Train a logistic regression model and compute predictions on the given test data."""
    if classifier == "SVM":
        model = select_svm(C, kernel)
    else:
        model = LogisticRegression(solver="liblinear", random_state=888, max_iter=5000, C=C)
    model.fit(
        train.x,
        train.y.to_numpy().ravel(),
        sample_weight=compute_weights(train)["instance weights"],
    )
    return pd.DataFrame(model.predict(test.x), columns=["preds"])



"""For assessing Calder-Verwer metric: 1-(P(Y=1|S=1)-P(Y=1|S!=1))."""

import pandas as pd

from ethicml.common import implements
from ethicml.metrics.prob_pos import ProbPos
from ethicml.utility.data_structures import DataTuple
from .metric import Metric


class CV(Metric):
    """Calder-Verwer."""

    @implements(Metric)
    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        from ethicml.evaluators.per_sensitive_attribute import (
            metric_per_sensitive_attribute,
            diff_per_sensitive_attribute,
        )

        per_sens = metric_per_sensitive_attribute(prediction, actual, ProbPos())
        diffs = diff_per_sensitive_attribute(per_sens)

        return 1 - list(diffs.values())[0]

    @property
    def name(self) -> str:
        """Getter for the metric name."""
        return "CV"

    @property
    def apply_per_sensitive(self) -> bool:
        """Can this metric be applied per sensitive attribute group?"""
        return False


class AbsCV(CV):
    """Absolute value of Calder-Verwer.

    This metric is supposed to make it easier to compare results.
    """

    @implements(Metric)
    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        cv_score = super().score(prediction, actual)
        # the following is equivalent to 1 - abs(diff)
        if cv_score > 1:
            return 2 - cv_score
        return cv_score

    @property
    def name(self) -> str:
        """Getter for the metric name."""
        return "CV absolute"



"""Implementation for Louizos et al Variational Fair Autoencoder."""
# pylint: disable=arguments-differ

from typing import List, Tuple
from torch import nn, Tensor


class Encoder(nn.Module):
    """Encoder for VFAE."""

    def __init__(
        self, enc_size: List[int], init_size: int, ld: int, activation: nn.Module = nn.ReLU()
    ) -> None:
        """Init Encoder."""
        super().__init__()
        self.encoder = nn.Sequential()
        if not enc_size:  # In the case that encoder size [] is specified
            self.z1_enc_mu = nn.Linear(init_size, ld)
            self.z1_enc_logvar = nn.Linear(init_size, ld)
        else:
            self.encoder.add_module("encoder layer 0", nn.Linear(init_size, enc_size[0]))
            if activation:
                self.encoder.add_module("encoder activation 0", activation)
            self.encoder.add_module("batch norm 0", nn.BatchNorm1d(enc_size[0]))
            for k in range(len(enc_size) - 1):
                self.encoder.add_module(
                    f"encoder layer {k+1}", nn.Linear(enc_size[k], enc_size[k + 1])
                )
                if activation:
                    self.encoder.add_module(f"encoder activation {k+1}", activation)
                self.encoder.add_module(
                    f"encoder batch norm {k+1}", nn.BatchNorm1d(enc_size[k + 1])
                )
            self.z1_enc_mu = nn.Linear(enc_size[-1], ld)
            self.z1_enc_logvar = nn.Linear(enc_size[-1], ld)

    def forward(self, input_: Tensor) -> Tuple[Tensor, Tensor]:  # type: ignore[override]
        """Forward pass of encoder."""
        step = self.encoder(input_)
        return self.z1_enc_mu(step), self.z1_enc_logvar(step)



"""Runs given metrics on given algorithms for given datasets."""
from pathlib import Path
from typing import List, Dict, Union, Sequence, Optional, NamedTuple
from collections import OrderedDict

import pandas as pd
from tqdm import tqdm

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithm
from ethicml.algorithms.postprocess.post_algorithm import PostAlgorithm
from ethicml.algorithms.preprocess.pre_algorithm import PreAlgorithm
from ethicml.utility.data_structures import DataTuple, TestTuple, TrainTestPair, Results
from .parallelism import run_in_parallel
from ..data.dataset import Dataset
from ..data.load import load_data
from .per_sensitive_attribute import (
    metric_per_sensitive_attribute,
    MetricNotApplicable,
    diff_per_sensitive_attribute,
    ratio_per_sensitive_attribute,
)
from ..metrics.metric import Metric
from ..preprocessing.train_test_split import DataSplitter, RandomSplit


def get_sensitive_combinations(metrics: List[Metric], train: DataTuple) -> List[str]:
    """Get all possible combinations of sensitive attribute and metrics."""
    poss_values: List[str] = []
    for col in train.s.columns:
        uniques = train.s[col].unique()
        for unique in uniques:
            poss_values.append(f"{col}_{unique}")

    return [f"{s}_{m.name}" for s in poss_values for m in metrics]


def per_sens_metrics_check(per_sens_metrics: Sequence[Metric]) -> None:
    """Check if the given metrics allow application per sensitive attribute."""
    for metric in per_sens_metrics:
        if not metric.apply_per_sensitive:
            raise MetricNotApplicable(
                f"Metric {metric.name} is not applicable per sensitive "
                f"attribute, apply to whole dataset instead"
            )


def run_metrics(
    predictions: pd.DataFrame,
    actual: DataTuple,
    metrics: Sequence[Metric] = (),
    per_sens_metrics: Sequence[Metric] = (),
) -> Dict[str, float]:
    """Run all the given metrics on the given predictions and return the results.

    Args:
        predictions: DataFrame with predictions
        actual: DataTuple with the labels
        metrics: list of metrics
        per_sens_metrics: list of metrics that are computed per sensitive attribute
    """
    result: Dict[str, float] = {}
    for metric in metrics:
        result[metric.name] = metric.score(predictions, actual)

    for metric in per_sens_metrics:
        per_sens = metric_per_sensitive_attribute(predictions, actual, metric)
        diff_per_sens = diff_per_sensitive_attribute(per_sens)
        ratio_per_sens = ratio_per_sensitive_attribute(per_sens)
        per_sens.update(diff_per_sens)
        per_sens.update(ratio_per_sens)
        for key, value in per_sens.items():
            result[f"{metric.name}_{key}"] = value
    return result  # SUGGESTION: we could return a DataFrame here instead of a dictionary


def load_results(
    dataset_name: str,
    transform_name: str,
    topic: Optional[str] = None,
    outdir: Path = Path(".") / "results",
) -> Optional[Results]:
    """Load results from a CSV file that was created by `evaluate_models`.

    Args:
        dataset_name: name of the dataset of the results
        transform_name: name of the transformation that was used for the results
        topic: (optional) topic string of the results
        outdir: directory where the results are stored

    Returns:
        DataFrame if the file exists; None otherwise
    """
    path_to_file = _result_path(outdir, dataset_name, transform_name, topic)
    return Results.from_file(path_to_file)


def _result_path(
    outdir: Path, dataset_name: str, transform_name: str, topic: Optional[str]
) -> Path:
    base_name: str = "" if topic is None else f"{topic}_"
    return outdir / f"{base_name}{dataset_name}_{transform_name}.csv"


def _delete_previous_results(
    outdir: Path, datasets: List[Dataset], transforms: Sequence[PreAlgorithm], topic: Optional[str]
) -> None:
    for dataset in datasets:
        transform_list = ["no_transform"]
        for preprocess_model in transforms:
            transform_list.append(preprocess_model.name)
        for transform_name in transform_list:
            path_to_file: Path = _result_path(outdir, dataset.name, transform_name, topic)
            if path_to_file.exists():
                path_to_file.unlink()


def evaluate_models(
    datasets: List[Dataset],
    preprocess_models: Sequence[PreAlgorithm] = (),
    inprocess_models: Sequence[InAlgorithm] = (),
    postprocess_models: Sequence[PostAlgorithm] = (),
    metrics: Sequence[Metric] = (),
    per_sens_metrics: Sequence[Metric] = (),
    repeats: int = 1,
    test_mode: bool = False,
    delete_prev: bool = False,
    splitter: Optional[DataSplitter] = None,
    topic: Optional[str] = None,
) -> Results:
    """Evaluate all the given models for all the given datasets and compute all the given metrics.

    Args:
        datasets: list of dataset objects
        preprocess_models: list of preprocess model objects
        inprocess_models: list of inprocess model objects
        postprocess_models: list of postprocess model objects
        metrics: list of metric objects
        per_sens_metrics: list of metric objects that will be evaluated per sensitive attribute
        repeats: number of repeats to perform for the experiments
        test_mode: if True, only use a small subset of the data so that the models run faster
        delete_prev:  False by default. If True, delete saved results in directory
        splitter: (optional) custom train-test splitter
        topic: (optional) a string that identifies the run; the string is prepended to the filename
    """
    # pylint: disable=too-many-arguments
    per_sens_metrics_check(per_sens_metrics)
    train_test_split: DataSplitter
    if splitter is None:
        train_test_split = RandomSplit(train_percentage=0.8, start_seed=0)
    else:
        train_test_split = splitter

    columns = ["dataset", "transform", "model", "split_id"]

    total_experiments = (
        len(datasets)
        * repeats
        * (len(preprocess_models) + ((1 + len(preprocess_models)) * len(inprocess_models)))
    )

    outdir = Path(".") / "results"
    outdir.mkdir(exist_ok=True)

    if delete_prev:
        _delete_previous_results(outdir, datasets, preprocess_models, topic)

    pbar = tqdm(total=total_experiments)
    for dataset in datasets:
        # ================================== begin: one repeat ====================================
        for split_id in range(repeats):
            train: DataTuple
            test: DataTuple
            train, test, split_info = train_test_split(load_data(dataset), split_id=split_id)
            if test_mode:
                # take smaller subset of training data to speed up training
                train = train.get_subset()

            to_operate_on: Dict[str, TrainTestPair] = {
                "no_transform": TrainTestPair(train=train, test=test)
            }

            # ========================== begin: run preprocessing models ==========================
            for pre_process_method in preprocess_models:
                logging: 'OrderedDict[str, str]' = OrderedDict()
                logging['model'] = pre_process_method.name
                logging['dataset'] = dataset.name
                logging['repeat'] = str(split_id)
                pbar.set_postfix(ordered_dict=logging)

                new_train, new_test = pre_process_method.run(train, test)
                to_operate_on[pre_process_method.name] = TrainTestPair(
                    train=new_train, test=new_test
                )

                pbar.update()
            # =========================== end: run preprocessing models ===========================

            # ========================= begin: loop over preprocessed data ========================
            for transform_name, transform in to_operate_on.items():

                transformed_train: DataTuple = transform.train
                transformed_test: Union[DataTuple, TestTuple] = transform.test
                results_df = pd.DataFrame(columns=columns)

                # ========================== begin: run inprocess models ==========================
                for model in inprocess_models:
                    logging = OrderedDict()
                    logging['model'] = model.name
                    logging['dataset'] = dataset.name
                    logging['transform'] = transform_name
                    logging['repeat'] = str(split_id)
                    pbar.set_postfix(ordered_dict=logging)

                    temp_res: Dict[str, Union[str, float]] = {
                        "dataset": dataset.name,
                        "transform": transform_name,
                        "model": model.name,
                        "split_id": split_id,
                        **split_info,
                    }

                    predictions: pd.DataFrame
                    predictions = model.run(transformed_train, transformed_test)

                    temp_res.update(run_metrics(predictions, test, metrics, per_sens_metrics))

                    for postprocess in postprocess_models:
                        # Post-processing has yet to be defined
                        # - leaving blank until we have an implementation to work with
                        pass

                    results_df = results_df.append(temp_res, ignore_index=True, sort=False)
                    pbar.update()
                # =========================== end: run inprocess models ===========================

                path_to_file = _result_path(outdir, dataset.name, transform_name, topic)
                results: Results = Results(results_df)
                # put old results before new results -> prepend=True
                results.append_from_file(path_to_file, prepend=True)
                results.save_as_csv(path_to_file)
            # ========================== end: loop over preprocessed data =========================
        # =================================== end: one repeat =====================================

    pbar.close()  # very important! when we're not using "with", we have to close tqdm manually

    preprocess_names = [model.name for model in preprocess_models]
    results = Results()  # create empty Results object
    for dataset in datasets:
        for transform_name in ["no_transform"] + preprocess_names:
            path_to_file = _result_path(outdir, dataset.name, transform_name, topic)
            results.append_from_file(path_to_file)
    return results


class _DataInfo(NamedTuple):
    test: DataTuple
    dataset_name: str
    transform_name: str
    split_info: Dict[str, float]


async def evaluate_models_async(
    datasets: List[Dataset],
    preprocess_models: Sequence[PreAlgorithm] = (),
    inprocess_models: Sequence[InAlgorithm] = (),
    postprocess_models: Sequence[PostAlgorithm] = (),
    metrics: Sequence[Metric] = (),
    per_sens_metrics: Sequence[Metric] = (),
    repeats: int = 1,
    test_mode: bool = False,
    delete_prev: bool = False,
    splitter: Optional[DataSplitter] = None,
    topic: Optional[str] = None,
    max_parallel: int = 1,
) -> Results:
    """Evaluate all the given models for all the given datasets and compute all the given metrics.

    Args:
        datasets: list of dataset objects
        preprocess_models: list of preprocess model objects
        inprocess_models: list of inprocess model objects
        postprocess_models: list of postprocess model objects
        metrics: list of metric objects
        per_sens_metrics: list of metric objects that will be evaluated per sensitive attribute
        repeats: number of repeats to perform for the experiments
        test_mode: if True, only use a small subset of the data so that the models run faster
        delete_prev:  False by default. If True, delete saved results in directory
        splitter: (optional) custom train-test splitter
        topic: (optional) a string that identifies the run; the string is prepended to the filename
        max_parallel: max number of threads ot run in parallel (default: 1)
    """
    # pylint: disable=too-many-arguments
    del postprocess_models  # not used at the moment
    per_sens_metrics_check(per_sens_metrics)
    train_test_split: DataSplitter
    if splitter is None:
        train_test_split = RandomSplit(train_percentage=0.8, start_seed=0)
    else:
        train_test_split = splitter

    default_transform_name = "no_transform"
    columns = ["dataset", "transform", "model", "split_id"]

    outdir = Path(".") / "results"  # OS-independent way of saying './results'
    outdir.mkdir(exist_ok=True)

    if delete_prev:
        _delete_previous_results(outdir, datasets, preprocess_models, topic)

    # ======================================= prepare data ========================================
    data_splits: List[TrainTestPair] = []
    test_data: List[_DataInfo] = []
    for dataset in datasets:
        for split_id in range(repeats):
            train: DataTuple
            test: DataTuple
            train, test, split_info = train_test_split(load_data(dataset), split_id=split_id)
            if test_mode:
                # take smaller subset of training data to speed up training
                train = train.get_subset()
            train = train.replace(name=f"{train.name} ({split_id})")
            data_splits.append(TrainTestPair(train, test))
            split_info.update({'split_id': split_id})
            test_data.append(_DataInfo(test, dataset.name, default_transform_name, split_info))

    # ===================================== preprocess models =====================================
    # run all preprocess models
    all_transformed = await run_in_parallel(preprocess_models, data_splits, max_parallel)

    # append the transformed data to `data_splits`
    for transformed, pre_model in zip(all_transformed, preprocess_models):
        for (transf_train, transf_test), data_info in zip(transformed, test_data):
            data_splits.append(TrainTestPair(transf_train, transf_test))
            test_data.append(
                _DataInfo(
                    data_info.test, data_info.dataset_name, pre_model.name, data_info.split_info
                )
            )

    # ===================================== inprocess models ======================================
    all_predictions = await run_in_parallel(inprocess_models, data_splits, max_parallel)

    # ====================================== compute metrics ======================================
    # transpose `all_results` so that the order in the results dataframe is correct
    num_cols = len(all_predictions[0])
    all_predictions_t = [[row[i] for row in all_predictions] for i in range(num_cols)]

    all_results = Results()

    # compute metrics, collect them and write them to files
    for preds_for_dataset, data_info in zip(all_predictions_t, test_data):
        # ============================= handle results of one dataset =============================
        results_df = pd.DataFrame(columns=columns)  # create empty results dataframe
        predictions: pd.DataFrame
        for predictions, model in zip(preds_for_dataset, inprocess_models):
            # construct a row of the results dataframe
            df_row: Dict[str, Union[str, float]] = {
                "dataset": data_info.dataset_name,
                "transform": data_info.transform_name,
                "model": model.name,
                **split_info,
            }
            df_row.update(run_metrics(predictions, data_info.test, metrics, per_sens_metrics))

            results_df = results_df.append(df_row, ignore_index=True, sort=False)

        # write results to CSV files and load previous results from the files if they already exist
        path_to_file = _result_path(outdir, data_info.dataset_name, data_info.transform_name, topic)
        results: Results = Results(results_df)
        # put old results before new results -> prepend=True
        results.append_from_file(path_to_file, prepend=True)
        results.save_as_csv(path_to_file)
        all_results.append_df(results.data)

    return all_results



"""EthicML tests"""


import pandas as pd

from ethicml.algorithms.inprocess import InAlgorithm, LR
from ethicml.algorithms.postprocess import Hardt, PostAlgorithm
from ethicml.utility import TrainTestPair, concat_tt

from .run_algorithm_test import count_true


def test_hardt(toy_train_test: TrainTestPair) -> None:
    """
    tests the hardt postprocessing technique
    Args:
        toy_train_test: Train-test pair of toy data
    """
    train, test = toy_train_test
    train_test = concat_tt([train, test], ignore_index=True)

    in_model: InAlgorithm = LR()
    assert in_model is not None
    assert in_model.name == "Logistic Regression, C=1.0"

    predictions: pd.DataFrame = in_model.run(train, train_test)

    # seperate out predictions on train set and predictions on test set
    pred_train = predictions.iloc[: train.y.shape[0]]
    pred_test = predictions.iloc[train.y.shape[0] :]
    assert count_true(pred_test.values == 1) == 211
    assert count_true(pred_test.values == -1) == 189

    post_model: PostAlgorithm = Hardt(-1, 1)
    fair_preds = post_model.run(pred_train, train, pred_test, test)
    assert count_true(fair_preds.values == 1) == 111
    assert count_true(fair_preds.values == -1) == 289



"""For assessing mean of logits."""

import pandas as pd

from ethicml.common import implements
from ethicml.utility.data_structures import DataTuple
from .metric import Metric


class ProbOutcome(Metric):
    """Mean of logits."""

    @implements(Metric)
    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        return prediction.to_numpy().sum() / prediction.size

    @property
    def name(self) -> str:
        """Getter for the metric name."""
        return "prob_outcome"



"""
Testing for plotting functionality
"""
from typing import Tuple, List

from matplotlib import pyplot as plt

import pytest

from ethicml.algorithms.inprocess import SVM, LR, Kamiran
from ethicml.algorithms.preprocess import Upsampler
from ethicml.evaluators import evaluate_models
from ethicml.metrics import Accuracy, CV, TPR, ProbPos, NMI
from ethicml.utility import DataTuple, Results
from ethicml.data import load_data, Adult, Toy
from ethicml.preprocessing import train_test_split
from ethicml.visualisation import save_2d_plot, save_label_plot, save_jointplot, plot_results
from tests.run_algorithm_test import get_train_test


@pytest.mark.usefixtures("plot_cleanup")  # fixtures are defined in `tests/conftest.py`
def test_plot():
    """test plot"""
    train, _ = get_train_test()
    save_2d_plot(train, "./plots/test.png")


@pytest.mark.usefixtures("plot_cleanup")
def test_joint_plot():
    """test joint plot"""
    train, _ = get_train_test()
    save_jointplot(train, "./plots/joint.png")


@pytest.mark.usefixtures("plot_cleanup")
def test_label_plot():
    """test label plot"""
    data: DataTuple = load_data(Adult())
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, _ = train_test

    save_label_plot(train, "./plots/labels.png")


@pytest.mark.usefixtures("plot_cleanup")
def test_plot_evals():
    """test plot evals"""
    results: Results = evaluate_models(
        datasets=[Adult(), Toy()],
        preprocess_models=[Upsampler(strategy="preferential")],
        inprocess_models=[LR(), SVM(kernel='linear'), Kamiran()],
        metrics=[Accuracy(), CV()],
        per_sens_metrics=[TPR(), ProbPos()],
        repeats=3,
        test_mode=True,
        delete_prev=True,
    )
    assert results.data['seed'][0] == results.data['seed'][1] == results.data['seed'][2] == 0
    assert results.data['seed'][3] == results.data['seed'][4] == results.data['seed'][5] == 2410
    assert results.data['seed'][6] == results.data['seed'][7] == results.data['seed'][8] == 4820

    figs_and_plots: List[Tuple[plt.Figure, plt.Axes]]

    # plot with metrics
    figs_and_plots = plot_results(results, Accuracy(), ProbPos())
    # num(datasets) * num(preprocess) * num(accuracy combinations) * num(prop_pos combinations)
    assert len(figs_and_plots) == 2 * 2 * 1 * 2

    # plot with column names
    figs_and_plots = plot_results(results, "Accuracy", "prob_pos_s_0")
    assert len(figs_and_plots) == 1 * 2 * 1 * 1

    with pytest.raises(ValueError, match='No matching columns found for Metric "NMI preds and y".'):
        plot_results(results, Accuracy(), NMI())

    with pytest.raises(ValueError, match='No column named "unknown metric".'):
        plot_results(results, "unknown metric", Accuracy())



"""Implementation of VFAE."""
from typing import Dict, List, Tuple

import pandas as pd
import torch
from torch import optim
from torch.utils.data import DataLoader
from torch.optim import Adam

from ethicml.data import Adult, Compas, Credit, German, NonBinaryToy, Sqf, Toy, Dataset
from ethicml.implementations.pytorch_common import CustomDataset, TestDataset
from ethicml.implementations.utils import PreAlgoArgs, load_data_from_flags, save_transformations
from ethicml.implementations.vfae_modules.utils import loss_function
from ethicml.implementations.vfae_modules.vfae_network import VFAENetwork
from ethicml.utility.data_structures import DataTuple, TestTuple


class VfaeArgs(PreAlgoArgs):
    """Args object of VFAE."""

    supervised: bool
    fairness: str
    batch_size: int
    epochs: int
    dataset: str
    z1_enc_size: List[int]
    z2_enc_size: List[int]
    z1_dec_size: List[int]


def get_dataset_obj_by_name(name: str) -> Dataset:
    """Given a dataset name, get the corresponding dataset object."""
    lookup: Dict[str, Dataset] = {
        "Adult": Adult(),
        "Compas": Compas(),
        "Credit": Credit(),
        "German": German(),
        "NonBinaryToy": NonBinaryToy(),
        "SQF": Sqf(),
        "Toy": Toy(),
    }

    if name not in lookup:
        raise NotImplementedError("That dataset doesn't exist")

    return lookup[name]


def train_and_transform(
    train: DataTuple, test: TestTuple, flags: VfaeArgs
) -> Tuple[DataTuple, TestTuple]:
    """Train the model and transform the dataset.

    Args:
        train:
        test:
        flags:

    Returns:
        Tuple of Encoded Train Dataset and Test Dataset.
    """
    dataset = get_dataset_obj_by_name(flags.dataset)

    # Set up the data
    train_data = CustomDataset(train)
    train_loader = DataLoader(train_data, batch_size=flags.batch_size)

    test_data = TestDataset(test)
    test_loader = DataLoader(test_data, batch_size=flags.batch_size)

    # Build Network
    model = VFAENetwork(
        dataset,
        flags.supervised,
        train_data.xdim,
        latent_dims=50,
        z1_enc_size=flags.z1_enc_size,
        z2_enc_size=flags.z2_enc_size,
        z1_dec_size=flags.z1_dec_size,
    ).to("cpu")
    optimizer = optim.Adam(model.parameters(), lr=1e-3)

    # Run Network
    for epoch in range(int(flags.epochs)):
        train_model(epoch, model, train_loader, optimizer, flags)

    # Transform output
    post_train: List[List[float]] = []
    post_test: List[List[float]] = []
    model.eval()
    with torch.no_grad():
        for _x, _s, _ in train_loader:
            z1_mu, z1_logvar = model.encode_z1(_x, _s)
            z1 = model.reparameterize(z1_mu, z1_logvar)
            post_train += z1.data.tolist()
        for _x, _s in test_loader:
            z1_mu, z1_logvar = model.encode_z1(_x, _s)
            z1 = model.reparameterize(z1_mu, z1_logvar)
            post_test += z1.data.tolist()

    return (
        DataTuple(x=pd.DataFrame(post_train), s=train.s, y=train.y, name=f"VFAE: {train.name}"),
        TestTuple(x=pd.DataFrame(post_test), s=test.s, name=f"VFAE: {test.name}"),
    )


def train_model(
    epoch: int, model: VFAENetwork, train_loader: DataLoader, optimizer: Adam, flags: VfaeArgs
) -> None:
    """Train the model.

    Args:
        epoch:
        model:
        train_loader:
        optimizer:
        flags:

    Returns:
        None
    """
    model.train()
    train_loss = 0
    for batch_idx, (data_x, data_s, data_y) in enumerate(train_loader):
        data_x = data_x.to("cpu")
        data_s = data_s.to("cpu")
        data_y = data_y.to("cpu")
        optimizer.zero_grad()
        z1_trip, z2_trip, z1_d_trip, x_dec, y_pred = model(data_x, data_s, data_y)

        data_trip = (data_x, data_s, data_y)

        loss_tuple = loss_function(flags, z1_trip, z2_trip, z1_d_trip, data_trip, x_dec, y_pred)
        prediction_loss, reconstruction_loss, kld_loss, mmd_loss = loss_tuple
        loss = kld_loss + reconstruction_loss + prediction_loss + mmd_loss
        loss.backward()
        # KLD_loss = KLD1 + KLD2
        train_loss += loss.item()
        optimizer.step()
        if batch_idx % 100 == 0:
            if flags.supervised:
                print(
                    f'train Epoch: {epoch} [{batch_idx * len(data_x)}/{len(train_loader.dataset)}'
                    f'({100. * batch_idx / len(train_loader):.0f}%)]\t'
                    f'Loss: {loss.item() / len(data_x):.6f}\t'
                    f'pred_loss: {prediction_loss.item():.6f}\t'
                    f'recon_loss: {reconstruction_loss.item():.6f}\t'
                    f'kld_loss: {kld_loss.item():.6f}\t'
                    f'mmd_loss: {flags.batch_size * mmd_loss.item():.6f}'
                )
            else:
                print(
                    f'train Epoch: {epoch} [{batch_idx * len(data_x)}/{len(train_loader.dataset)}'
                    f'({100. * batch_idx / len(train_loader):.0f}%)]\t'
                    f'Loss: {loss.item() / len(data_x):.6f}\t'
                    f'recon_loss: {reconstruction_loss.item():.6f}\t'
                    f'mmd_loss: {flags.batch_size * mmd_loss.item():.6f}'
                )

    print(f'====> Epoch: {epoch} Average loss: {train_loss / len(train_loader.dataset):.4f}')


def main():
    """Main method to run model."""
    args = VfaeArgs(explicit_bool=True).parse_args()
    train, test = load_data_from_flags(args)
    save_transformations(train_and_transform(train, test, args), args)


if __name__ == "__main__":
    main()
